import json

import pytest
import responses

from datarobot import Model
from datarobot.errors import ClientError
from datarobot.models.residuals import ResidualsChart


@pytest.fixture
def residuals_chart_example_data():
    return {
        "coefficientOfDetermination": 0.9879402133296861,
        "data": [
            [148.38722636305582, 136],
            [235.6488840730197, 234],
            [144.61655369382572, 145],
            [116.52478603021166, 114],
            [219.86150357121636, 233],
            [264.53947412968864, 264],
            [186.07665522690363, 184],
            [472.31241652697463, 465],
            [481.37146812597763, 491],
            [122.96392805756432, 119],
            [320.0423854913889, 348],
            [578.6310089767073, 622],
            [497.75908894718265, 508],
            [329.750738665105, 340],
            [116.34834803467115, 118],
            [314.76117436560696, 306],
            [207.70754892487585, 188],
            [222.64967210517932, 229],
            [307.73869034965304, 305],
            [154.71230780244468, 149],
            [187.51570240220502, 181],
            [258.37348812623134, 259],
            [340.7130549526243, 315],
        ],
        "residualMean": 1.0432128285950566,
    }


@pytest.fixture
def residuals_chart_validation_data(model_id, residuals_chart_example_data):
    return dict(source="validation", sourceModelId=model_id, **residuals_chart_example_data)


@pytest.fixture
def residuals_chart_validation_api_response(residuals_chart_example_data):
    return {"residuals": {"validation": residuals_chart_example_data}}


@pytest.fixture
def residuals_chart_parent_model_validation_data(parent_model_id, residuals_chart_example_data):
    return dict(source="validation", sourceModelId=parent_model_id, **residuals_chart_example_data)


@pytest.fixture
def residuals_chart_parent_model_validation_api_response(residuals_chart_validation_api_response):
    return residuals_chart_validation_api_response


@pytest.fixture
def residuals_chart_parent_model_holdout_data(parent_model_id, residuals_chart_example_data):
    return dict(source="holdout", sourceModelId=parent_model_id, **residuals_chart_example_data)


@pytest.fixture
def residuals_chart_parent_model_holdout_api_response(residuals_chart_example_data):
    return {"residuals": {"holdout": residuals_chart_example_data}}


def test_instantiation(residuals_chart_validation_data):
    residuals_chart = ResidualsChart.from_server_data(residuals_chart_validation_data)

    assert residuals_chart.source == residuals_chart_validation_data["source"]
    assert residuals_chart.data == residuals_chart_validation_data["data"]
    assert (
        residuals_chart.coefficient_of_determination
        == residuals_chart_validation_data["coefficientOfDetermination"]
    )
    assert residuals_chart.residual_mean == residuals_chart_validation_data["residualMean"]
    assert residuals_chart.standard_deviation is None


def test_instantiation_with_std_deviation(residuals_chart_validation_data):
    residuals_chart_validation_data["standard_deviation"] = 1.1111
    residuals_chart = ResidualsChart.from_server_data(residuals_chart_validation_data)

    assert residuals_chart.source == residuals_chart_validation_data["source"]
    assert residuals_chart.data == residuals_chart_validation_data["data"]
    assert (
        residuals_chart.coefficient_of_determination
        == residuals_chart_validation_data["coefficientOfDetermination"]
    )
    assert residuals_chart.residual_mean == residuals_chart_validation_data["residualMean"]
    assert (
        residuals_chart.standard_deviation == residuals_chart_validation_data["standard_deviation"]
    )


def test_future_proof(residuals_chart_validation_data):
    data_with_future_keys = dict(residuals_chart_validation_data, new_key="some future lift data")
    data_with_future_keys["new_key"] = "some future data"
    ResidualsChart.from_server_data(data_with_future_keys)


@pytest.fixture
def residuals_chart_validation_data_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/residuals/validation/".format(
        project_id, model_id
    )


@pytest.fixture
def residuals_chart_parent_model_validation_data_url(project_id, parent_model_id):
    return "https://host_name.com/projects/{}/models/{}/residuals/validation/".format(
        project_id, parent_model_id
    )


@pytest.fixture
def residuals_chart_list_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/residuals/".format(project_id, model_id)


@pytest.fixture
def residuals_chart_parent_list_url(project_id, parent_model_id):
    return "https://host_name.com/projects/{}/models/{}/residuals/".format(
        project_id, parent_model_id
    )


@pytest.fixture
def frozen_model_url(project_id, model_id):
    return "https://host_name.com/projects/{}/frozenModels/{}/".format(project_id, model_id)


@responses.activate
def test_get_validation_residuals_chart(
    residuals_chart_validation_api_response,
    residuals_chart_validation_data,
    residuals_chart_validation_data_url,
    project_id,
    model_id,
):
    responses.add(
        responses.GET,
        residuals_chart_validation_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(residuals_chart_validation_api_response),
    )
    model = Model(id=model_id, project_id=project_id)
    residuals_chart = model.get_residuals_chart("validation")

    assert residuals_chart.source == residuals_chart_validation_data["source"]
    assert residuals_chart.data == residuals_chart_validation_data["data"]
    assert (
        residuals_chart.coefficient_of_determination
        == residuals_chart_validation_data["coefficientOfDetermination"]
    )
    assert residuals_chart.residual_mean == residuals_chart_validation_data["residualMean"]


@responses.activate
def test_get_frozen_validation_residuals_chart_no_fallback(
    residuals_chart_validation_data_url, project_id, model_id
):
    responses.add(responses.GET, residuals_chart_validation_data_url, status=404)
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    with pytest.raises(ClientError):
        model.get_residuals_chart("validation")


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_frozen_validation_residuals_chart_with_fallback(
    residuals_chart_parent_model_validation_data,
    residuals_chart_parent_model_validation_api_response,
    residuals_chart_validation_data_url,
    residuals_chart_parent_model_validation_data_url,
    frozen_model_url,
    project_id,
    model_id,
    parent_model_id,
    frozen_json,
):

    responses.add(responses.GET, residuals_chart_validation_data_url, status=404)
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        residuals_chart_parent_model_validation_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(residuals_chart_parent_model_validation_api_response),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    residuals_chart = model.get_residuals_chart("validation", fallback_to_parent_insights=True)

    assert residuals_chart.source == residuals_chart_parent_model_validation_data["source"]
    assert residuals_chart.data == residuals_chart_parent_model_validation_data["data"]
    assert residuals_chart.coefficient_of_determination == (
        residuals_chart_parent_model_validation_data["coefficientOfDetermination"]
    )
    assert (
        residuals_chart.residual_mean
        == residuals_chart_parent_model_validation_data["residualMean"]
    )


@responses.activate
def test_get_all_residuals_charts(
    residuals_chart_validation_data,
    residuals_chart_validation_api_response,
    residuals_chart_list_url,
    project_id,
    model_id,
):
    responses.add(
        responses.GET,
        residuals_chart_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps(residuals_chart_validation_api_response),
    )
    model = Model(id=model_id, project_id=project_id)
    residuals_chart_list = model.get_all_residuals_charts()

    assert len(residuals_chart_list) == 1
    residuals_chart = residuals_chart_list[0]
    assert residuals_chart.source == residuals_chart_validation_data["source"]
    assert residuals_chart.data == residuals_chart_validation_data["data"]
    assert (
        residuals_chart.coefficient_of_determination
        == residuals_chart_validation_data["coefficientOfDetermination"]
    )
    assert residuals_chart.residual_mean == residuals_chart_validation_data["residualMean"]


@responses.activate
def test_get_frozen_all_residuals_charts_no_fallback(
    residuals_chart_validation_data,
    residuals_chart_list_url,
    project_id,
    model_id,
    residuals_chart_validation_api_response,
):
    responses.add(
        responses.GET,
        residuals_chart_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps(residuals_chart_validation_api_response),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    residuals_chart_list = model.get_all_residuals_charts()

    assert len(residuals_chart_list) == 1
    residuals_chart = residuals_chart_list[0]
    assert residuals_chart.source == residuals_chart_validation_data["source"]
    assert residuals_chart.data == residuals_chart_validation_data["data"]
    assert (
        residuals_chart.coefficient_of_determination
        == residuals_chart_validation_data["coefficientOfDetermination"]
    )
    assert residuals_chart.residual_mean == residuals_chart_validation_data["residualMean"]


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_frozen_all_residuals_charts_with_fallback(
    residuals_chart_validation_data,
    residuals_chart_validation_api_response,
    residuals_chart_parent_model_validation_data,
    residuals_chart_parent_model_holdout_data,
    residuals_chart_parent_model_validation_api_response,
    residuals_chart_parent_model_holdout_api_response,
    residuals_chart_list_url,
    residuals_chart_parent_list_url,
    frozen_model_url,
    project_id,
    model_id,
    parent_model_id,
    frozen_json,
):
    responses.add(
        responses.GET,
        residuals_chart_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps(residuals_chart_validation_api_response),
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type="application/json",
    )
    parent_list_response = {
        "residuals": residuals_chart_parent_model_validation_api_response["residuals"].copy()
    }
    parent_list_response["residuals"][
        "holdout"
    ] = residuals_chart_parent_model_holdout_api_response["residuals"]["holdout"]
    responses.add(
        responses.GET,
        residuals_chart_parent_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps(parent_list_response),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    residuals_chart_list = model.get_all_residuals_charts(fallback_to_parent_insights=True)

    assert len(residuals_chart_list) == 2
    residuals_chart = residuals_chart_list[0]
    assert residuals_chart.source == residuals_chart_validation_data["source"]
    assert residuals_chart.data == residuals_chart_validation_data["data"]
    assert (
        residuals_chart.coefficient_of_determination
        == residuals_chart_validation_data["coefficientOfDetermination"]
    )
    assert residuals_chart.residual_mean == residuals_chart_validation_data["residualMean"]

    residuals_chart = residuals_chart_list[1]
    assert residuals_chart.source == residuals_chart_parent_model_holdout_data["source"]
    assert residuals_chart.data == residuals_chart_parent_model_holdout_data["data"]
    assert residuals_chart.coefficient_of_determination == (
        residuals_chart_parent_model_holdout_data["coefficientOfDetermination"]
    )
    assert (
        residuals_chart.residual_mean == residuals_chart_parent_model_holdout_data["residualMean"]
    )



from mock import patch
import pytest
import responses

from datarobot import ExternalLiftChart
from datarobot.errors import ClientError
from datarobot.utils import from_api


class TestExternalLiftChart(object):
    @staticmethod
    def create_response_data(lift_chart_bins_data, count):
        return [
            {"bins": lift_chart_bins_data, "datasetId": "dataset_id_{}".format(i)}
            for i in range(count)
        ]

    @pytest.fixture
    def url(self, project_id, model_id):
        return "https://host_name.com/projects/{}/models/{}/datasetLiftCharts/".format(
            project_id, model_id
        )

    def test_instantiation(self, lift_chart_bins_data):
        one_lift_chart = self.create_response_data(lift_chart_bins_data, 1)[0]
        lc = ExternalLiftChart.from_server_data(one_lift_chart)

        assert lc.dataset_id == one_lift_chart["datasetId"]
        assert lc.bins == from_api(lift_chart_bins_data)

    def test_future_proof(self, lift_chart_bins_data):
        one_lift_chart = self.create_response_data(lift_chart_bins_data, 1)[0]
        data_with_future_keys = dict(one_lift_chart, new_key="some future lift data")
        data_with_future_keys["bins"][0]["new_key"] = "some future bin data"
        ExternalLiftChart.from_server_data(data_with_future_keys)

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights.external_lift_chart."
        "DEFAULT_BATCH_SIZE",
        10,
    )
    def test_list_lift_chart_all_results(self, lift_chart_bins_data, url, project_id, model_id):
        url_page1 = url + "?limit=10&offset=0"
        url_page2 = url + "?limit=10&offset=10"
        all_data = self.create_response_data(lift_chart_bins_data, 11)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[10:])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True
        )

        lift_charts = ExternalLiftChart.list(project_id, model_id, limit=0)
        assert len(lift_charts) == 11
        for i, chart in enumerate(lift_charts):
            assert isinstance(chart, ExternalLiftChart)
            assert chart.dataset_id == "dataset_id_{}".format(i)
            assert chart.bins == from_api(lift_chart_bins_data)

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights.external_lift_chart."
        "DEFAULT_BATCH_SIZE",
        10,
    )
    def test_list_lift_chart_all_results_with_offset(
        self, lift_chart_bins_data, url, project_id, model_id
    ):
        url_page1 = url + "?limit=10&offset=2"
        url_page2 = url + "?limit=10&offset=12"
        all_data = self.create_response_data(lift_chart_bins_data, 15)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[2:12])
        resp_data_2 = dict(count=3, previous=url_page1, next=None, data=all_data[12:])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True
        )

        lift_charts = ExternalLiftChart.list(project_id, model_id, limit=0, offset=2)
        assert len(lift_charts) == 13
        assert lift_charts[0].dataset_id == "dataset_id_2"

    @responses.activate
    def test_list_lift_chart_limited(self, lift_chart_bins_data, url, project_id, model_id):
        url_page = url + "?limit=3&offset=2"
        all_data = self.create_response_data(lift_chart_bins_data, 5)
        resp_data = dict(count=3, previous=None, next=None, data=all_data[2:5])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        lift_charts = ExternalLiftChart.list(project_id, model_id, limit=3, offset=2)
        assert len(lift_charts) == 3
        for i, chart in enumerate(lift_charts):
            assert isinstance(chart, ExternalLiftChart)
            assert chart.dataset_id == "dataset_id_{}".format(i + 2)
            assert chart.bins == from_api(lift_chart_bins_data)

    @responses.activate
    def test_get_lift_chart_success(self, lift_chart_bins_data, url, project_id, model_id):
        dataset_id = "dataset_id_0"
        url_page = url + "?datasetId={}&limit=100&offset=0".format(dataset_id)
        all_data = self.create_response_data(lift_chart_bins_data, 2)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        chart = ExternalLiftChart.get(project_id, model_id, dataset_id)
        assert isinstance(chart, ExternalLiftChart)
        assert chart.dataset_id == "dataset_id_0"
        assert chart.bins == from_api(lift_chart_bins_data)

    @responses.activate
    def test_get_lift_chart_failure(self, url, project_id, model_id):
        dataset_id = "dataset_id_0"
        url_page = url + "?datasetId={}&limit=100&offset=0".format(dataset_id)
        resp_data = dict(count=0, previous=None, next=None, data=[])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        with pytest.raises(ClientError, match="Requested lift chart does not exist."):
            ExternalLiftChart.get(project_id, model_id, dataset_id)



from collections import OrderedDict

import dateutil
import pandas as pd
import trafaret as t

from datarobot.models.api_object import APIObject

from ..enums import ACCURACY_METRIC
from ..helpers.deployment_monitoring import DeploymentQueryBuilderMixin
from ..utils import encode_utf8_if_py2, from_api


class Accuracy(APIObject, DeploymentQueryBuilderMixin):
    """Deployment accuracy information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve accuracy metrics
    period : dict
        the time period used to retrieve accuracy metrics
    metrics : dict
        the accuracy metrics
    """

    _path = "deployments/{}/accuracy/"
    _period = t.Dict(
        {
            t.Key("start"): t.String >> (lambda s: dateutil.parser.parse(s)),
            t.Key("end"): t.String >> (lambda s: dateutil.parser.parse(s)),
        }
    )
    _metric = t.Dict(
        {
            t.Key("value", optional=True): t.Float | t.Null,
            t.Key("baseline_value", optional=True): t.Float | t.Null,
            t.Key("percent_change", optional=True): t.Float | t.Null,
        }
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("period"): _period,
            t.Key("metrics"): t.Mapping(t.Enum(*ACCURACY_METRIC.ALL), _metric),
            t.Key("model_id"): t.String() | t.Null,
        }
    ).allow_extra("*")

    def __init__(self, period=None, metrics=None, model_id=None):
        self.period = period or {}
        self.metrics = metrics or {}
        self.model_id = model_id

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}({} | {} - {})".format(
                self.__class__.__name__,
                self.model_id,
                self.period.get("start"),
                self.period.get("end"),
            )
        )

    def __getitem__(self, item):
        return self.metrics.get(item, {}).get("value")

    @classmethod
    def get(cls, deployment_id, model_id=None, start_time=None, end_time=None):
        """Retrieve values of accuracy metrics over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period

        Returns
        -------
        accuracy : Accuracy
            the queried accuracy metrics information

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, Accuracy
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            accuracy = Accuracy.get(deployment.id)
            accuracy.period['end']
            >>>'2019-08-01 00:00:00+00:00'
            accuracy.metric['LogLoss']['value']
            >>>0.7533
            accuracy.metric_values['LogLoss']
            >>>0.7533
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time, end_time=end_time, model_id=model_id
        )
        data = cls._client.get(path, params=params).json()

        # we don't want to convert keys of the metrics object
        metrics = data.pop("metrics")
        for metric, value in metrics.items():
            metrics[metric] = from_api(value, keep_null_keys=True)

        data = from_api(data, keep_null_keys=True)
        data["metrics"] = metrics
        return cls.from_data(data)

    @property
    def metric_values(self):
        """The value for all metrics, keyed by metric name.

        Returns
        -------
        metric_values: OrderedDict
        """

        return {name: value.get("value") for name, value in self.metrics.items()}

    @property
    def metric_baselines(self):
        """The baseline value for all metrics, keyed by metric name.

        Returns
        -------
        metric_baselines: OrderedDict
        """

        return {name: value.get("baseline_value") for name, value in self.metrics.items()}

    @property
    def percent_changes(self):
        """The percent change of value over baseline for all metrics, keyed by metric name.

        Returns
        -------
        percent_changes: OrderedDict
        """

        return {name: value.get("percent_change") for name, value in self.metrics.items()}


class AccuracyOverTime(APIObject, DeploymentQueryBuilderMixin):
    """Deployment accuracy over time information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve accuracy metric
    metric : str
        the accuracy metric being retrieved
    buckets : dict
        how the accuracy metric changes over time
    summary : dict
        summary for the accuracy metric
    baseline : dict
        baseline for the accuracy metric
    """

    _path = "deployments/{}/accuracyOverTime/"
    _period = t.Dict(
        {
            t.Key("start"): t.String >> (lambda s: dateutil.parser.parse(s)),
            t.Key("end"): t.String >> (lambda s: dateutil.parser.parse(s)),
        }
    )
    _bucket = t.Dict(
        {
            t.Key("period"): _period | t.Null,
            t.Key("value"): t.Float | t.Null,
            t.Key("sample_size"): t.Int | t.Null,
        }
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("buckets"): t.List(_bucket),
            t.Key("summary"): _bucket,
            t.Key("baseline"): _bucket,
            t.Key("metric"): t.String(),
            t.Key("model_id"): t.String() | t.Null,
        }
    ).allow_extra("*")

    def __init__(self, buckets=None, summary=None, baseline=None, metric=None, model_id=None):
        self.buckets = buckets
        self.summary = summary
        self.baseline = baseline
        self.metric = metric
        self.model_id = model_id

    def __repr__(self):
        period = self.summary.get("period") or {}
        return encode_utf8_if_py2(
            u"{}({} | {} | {} - {})".format(
                self.__class__.__name__,
                self.model_id,
                self.metric,
                period.get("start"),
                period.get("end"),
            )
        )

    @classmethod
    def get(
        cls,
        deployment_id,
        metric=None,
        model_id=None,
        start_time=None,
        end_time=None,
        bucket_size=None,
    ):
        """Retrieve information about how an accuracy metric changes over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        metric : ACCURACY_METRIC
            the accuracy metric to retrieve
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        bucket_size : str
            time duration of a bucket, in ISO 8601 time duration format

        Returns
        -------
        accuracy_over_time : AccuracyOverTime
            the queried accuracy metric over time information

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, AccuracyOverTime
            from datarobot.enums import ACCURACY_METRICS
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            accuracy_over_time = AccuracyOverTime.get(deployment.id, metric=ACCURACY_METRIC.LOGLOSS)
            accuracy_over_time.metric
            >>>'LogLoss'
            accuracy_over_time.metric_values
            >>>{datetime.datetime(2019, 8, 1): 0.73, datetime.datetime(2019, 8, 2): 0.55}
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            metric=metric,
            bucket_size=bucket_size,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)

    @classmethod
    def get_as_dataframe(
        cls, deployment_id, metrics, model_id=None, start_time=None, end_time=None, bucket_size=None
    ):
        """Retrieve information about how a list of accuracy metrics change over
        a certain time period as pandas DataFrame.

        In the returned DataFrame, the columns corresponds to the metrics being retrieved;
        the rows are labeled with the start time of each bucket.

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        metrics : [ACCURACY_METRIC]
            the accuracy metrics to retrieve
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        bucket_size : str
            time duration of a bucket, in ISO 8601 time duration format

        Returns
        -------
        accuracy_over_time: pd.DataFrame
        """

        if metrics is None:
            metrics = []
        metric_names = []
        metric_dataframes = []
        for metric_name in metrics:
            fetched = AccuracyOverTime.get(
                deployment_id,
                model_id=model_id,
                metric=metric_name,
                start_time=start_time,
                end_time=end_time,
                bucket_size=bucket_size,
            )
            dataframe = pd.DataFrame.from_dict(fetched.bucket_values, orient="index")
            metric_names.append(fetched.metric)
            metric_dataframes.append(dataframe)
        combined = pd.concat(metric_dataframes, axis="columns")
        combined.columns = metric_names
        return combined

    @property
    def bucket_values(self):
        """The metric value for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_values: OrderedDict
        """

        values = [(bucket["period"]["start"], bucket["value"]) for bucket in self.buckets]
        return OrderedDict(values)

    @property
    def bucket_sample_sizes(self):
        """The sample size for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_sample_sizes: OrderedDict
        """

        values = [(bucket["period"]["start"], bucket["sample_size"]) for bucket in self.buckets]
        return OrderedDict(values)



import csv
import io
import json
import tempfile

import pandas as pd
import pytest
import responses
import six
import trafaret

from datarobot import TrainingPredictions
from datarobot.models.training_predictions import TrainingPredictionsRow


def _write_csv(lines):

    if six.PY2:

        buf = io.BytesIO()
        wr = csv.writer(buf)
        wr.writerows(lines)
        return buf.getvalue()

    buf = io.StringIO()
    wr = csv.writer(buf)
    wr.writerows(lines)
    return buf.getvalue().encode("utf-8")


@pytest.fixture
def prediction_id():
    return "59d92fe8962d7464ca8c6bd6"


@pytest.fixture
def predictions_with_shap_exp_csv():
    return """row_id,partition_id,prediction,class_1.0,class_0.0,explained_class,Explanation_1_feature_name,Explanation_1_feature_value,Explanation_1_strength,shap_remaining_total,shap_base_value
        0,TrainTestOverlap,0.0,0.311178247734139,0.6888217522658611,1.0,number_diagnoses,9,-0.07643818895488698,0.006562548084582246,0.39640626311300003
        1,TrainTestOverlap,0.0,0.2829861111111111,0.7170138888888888,1.0,number_diagnoses,6,0.06648347981438553,-0.013534302788801085,0.39640626311300003
        2,TrainTestOverlap,0.0,0.42699724517906334,0.5730027548209367,1.0,number_diagnoses,9,0.08556893050818416,-0.032619753482599656,0.39640626311300003
        3,TrainTestOverlap,0.0,0.32653061224489793,0.6734693877551021,1.0,number_diagnoses,3,0.05807900167179779,-0.0051298246462132885,0.39640626311300003
        4,TrainTestOverlap,0.0,0.06,0.94,1.0,number_inpatient,0,-0.056383896138088974,-0.0570362457984236,0.39640626311300003
        5,TrainTestOverlap,0.0,0.3774597495527728,0.6225402504472273,1.0,number_diagnoses,8,0.04847090471521205,-0.03250602234256627,0.39640626311300003
        6,TrainTestOverlap,0.0,0.13768115942028986,0.8623188405797102,1.0,admission_type_id,Elective,0.08996873371035077,0.040252262429918834,0.39640626311300003
        7,TrainTestOverlap,1.0,0.5300353356890459,0.4699646643109541,1.0,admission_type_id,Emergency,0.16263335056706338,-0.03241235442679378,0.39640626311300003
        8,TrainTestOverlap,1.0,0.6408668730650154,0.35913312693498456,1.0,number_inpatient,1,0.34801989552094514,-0.055934082238188985,0.39640626311300003
        9,TrainTestOverlap,0.0,0.2027972027972028,0.7972027972027972,1.0,number_inpatient,0,0.1146691220002886,-0.08407812406711129,0.39640626311300003
    """


@responses.activate
def test_get__no_requests_made(project_id, prediction_id):
    TrainingPredictions.get(project_id, prediction_id)
    assert len(responses.calls) == 0


@responses.activate
def test_list__ok(project_id, model_id, prediction_id):
    training_prediction_stub = [
        {
            "url": "https://host_name.com/projects/{}/trainingPredictions/{}//".format(
                project_id, prediction_id
            ),
            "id": prediction_id,
            "modelId": model_id,
            "dataSubset": "all",
        },
        {
            "url": "https://host_name.com/projects/{}/trainingPredictions/{}//".format(
                project_id, prediction_id
            ),
            "id": prediction_id,
            "modelId": model_id,
            "dataSubset": "all",
            "explanationAlgorithm": "shap",
            "maxExplanations": 7,
            "shapWarnings": {
                "partitionName": "holdout",
                "value": {"mismatchRowCount": 1, "maxNormalizedMismatch": 0.01},
            },
        },
    ]
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/trainingPredictions/".format(project_id),
        status=200,
        body=json.dumps({"data": training_prediction_stub}),
    )

    training_predictions_list = TrainingPredictions.list(project_id)

    assert len(training_predictions_list) == 2
    training_prediction = training_predictions_list[0]
    assert training_prediction.project_id == project_id
    assert training_prediction.model_id == model_id
    assert training_prediction.prediction_id == prediction_id
    training_prediction_with_shap_exp = training_predictions_list[1]
    assert training_prediction_with_shap_exp.project_id == project_id
    assert training_prediction_with_shap_exp.model_id == model_id
    assert training_prediction_with_shap_exp.explanation_algorithm == "shap"
    assert training_prediction_with_shap_exp.max_explanations == 7
    assert training_prediction_with_shap_exp.shap_warnings == {
        "partition_name": "holdout",
        "value": {"mismatch_row_count": 1, "max_normalized_mismatch": 0.01},
    }


@responses.activate
def test_list__no_url_in_api_response__error(project_id, model_id, prediction_id):
    training_prediction_stub = {
        "url": "",
        "id": prediction_id,
        "modelId": model_id,
        "trainingPredictionSet": "all",
    }
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/trainingPredictions/".format(project_id),
        status=200,
        body=json.dumps({"data": [training_prediction_stub]}),
    )

    with pytest.raises(trafaret.DataError):
        TrainingPredictions.list(project_id)


@responses.activate
def test_iterate_rows__empty_predictions__ok(project_id, prediction_id):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    response = {"data": [], "next": "{}?offset=100&limit=100".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(response), status=200, content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)

    assert list(obj.iterate_rows()) == []


@responses.activate
def test_iterate_rows__malfromed_predictions__data_error(project_id, prediction_id):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    items = [
        {
            "row_id": "wrong-id",
            "partition_id": "holdout",
            "prediction": 0.42,
            "prediction_values": [],
        }
    ]
    response = {"data": items, "next": "{}?offset=100&limit=100".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(response), status=200, content_type="application/json",
    )

    with pytest.raises(trafaret.DataError):
        obj = TrainingPredictions.get(project_id, prediction_id)
        it = obj.iterate_rows()
        next(it)


@responses.activate
def test_iterate_rows__extra_fields__ignored(project_id, prediction_id):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    items = [
        {
            "row_id": 1,
            "partition_id": 2,
            "prediction": 0.42,
            "prediction_values": [
                {
                    "label": "test-label",
                    "value": 0.42,
                    "extra-prediction": "you are going to win a lottery for sure",
                }
            ],
        },
    ]
    response = {"data": items, "next": "{}?offset=100&limit=100".format(url), "prev": None}
    responses.add(
        responses.GET, url, body=json.dumps(response), status=200, content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    assert list(obj.iterate_rows()) == [
        TrainingPredictionsRow(
            row_id=1,
            partition_id=2,
            prediction=0.42,
            prediction_values=[{"label": "test-label", "value": 0.42}],
            timestamp=None,
            forecast_point=None,
            forecast_distance=None,
            series_id=None,
            prediction_explanations=None,
            shap_metadata=None,
        )
    ]


@responses.activate
def test_iterate_rows__all_rows_at_once__ok(project_id, prediction_id):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    items = [
        {"row_id": 1, "partition_id": 2, "prediction": 0.42, "prediction_values": []},
        {"row_id": 2, "partition_id": 2, "prediction": 0.22, "prediction_values": []},
        {"row_id": 3, "partition_id": 2, "prediction": 0.82, "prediction_values": []},
    ]
    response = {"data": items, "next": None, "prev": None}
    responses.add(
        responses.GET, url, body=json.dumps(response), status=200, content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    assert len(list(obj.iterate_rows(batch_size=3))) == 3


@responses.activate
def test_iterate_rows__one_prediction__ok(project_id, prediction_id):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    items = [
        {"row_id": 1, "partition_id": 2, "prediction": 0.42, "prediction_values": []},
    ]
    response = {"data": items, "next": "{}?offset=100&limit=100".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(response), status=200, content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)

    assert list(obj.iterate_rows()) == [
        TrainingPredictionsRow(
            row_id=1,
            partition_id=2,
            prediction=0.42,
            prediction_values=[],
            timestamp=None,
            forecast_point=None,
            forecast_distance=None,
            series_id=None,
            prediction_explanations=None,
            shap_metadata=None,
        ),
    ]
    assert len(responses.calls) == 1


@responses.activate
def test_iterate_rows__one_prediction_with_shap_exp(project_id, prediction_id):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    items = [
        {
            "row_id": 1,
            "partition_id": 2,
            "prediction": 0.42,
            "prediction_values": [],
            "prediction_explanations": [
                {
                    "feature": "first_name",
                    "feature_value": "some_value",
                    "strength": -0.07,
                    "label": 1.0,
                },
                {"feature": "second_name", "feature_value": "0", "strength": -0.05, "label": 1.0},
            ],
            "shap_metadata": {
                "shap_base_value": 0.39,
                "shap_remaining_total": 0.03,
                "warnings": {"mismatch_row_count": 1, "max_normalized_mismatch": 0.01},
            },
        },
    ]
    response = {"data": items, "next": "{}?offset=100&limit=100".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(response), status=200, content_type="application/json"
    )

    training_predictions = TrainingPredictions.get(project_id, prediction_id)

    data = list(training_predictions.iterate_rows())
    assert len(data) == 1
    training_prediction = data[0]
    assert training_prediction.row_id == 1
    assert training_prediction.partition_id == 2
    assert training_prediction.prediction == 0.42
    assert training_prediction.prediction_values == []
    assert training_prediction.timestamp is None
    assert training_prediction.forecast_point is None
    assert training_prediction.forecast_distance is None
    assert training_prediction.series_id is None
    assert training_prediction.series_id is None
    assert training_prediction.prediction_explanations == [
        {"feature": "first_name", "feature_value": "some_value", "strength": -0.07, "label": 1.0},
        {"feature": "second_name", "feature_value": "0", "strength": -0.05, "label": 1.0},
    ]
    assert training_prediction.shap_metadata == {
        "shap_base_value": 0.39,
        "shap_remaining_total": 0.03,
        "warnings": {"mismatch_row_count": 1, "max_normalized_mismatch": 0.01},
    }
    assert len(responses.calls) == 1


@responses.activate
def test_iterate_rows__two_predictions_with_limit_equals_one__three_requests_were_made(
    project_id, prediction_id,
):
    items = [
        {"row_id": 1, "partition_id": 2, "prediction": 0.42, "prediction_values": []},
        {"row_id": 2, "partition_id": 2, "prediction": 0.22, "prediction_values": []},
    ]
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    body = {"data": items[:1], "next": "{}?offset=1&limit=1".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(body), status=200, content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)

    it = obj.iterate_rows(batch_size=1)

    assert next(it, None) == TrainingPredictionsRow(
        row_id=1,
        partition_id=2,
        prediction=0.42,
        prediction_values=[],
        timestamp=None,
        forecast_point=None,
        forecast_distance=None,
        series_id=None,
        prediction_explanations=None,
        shap_metadata=None,
    )
    assert len(responses.calls) == 1

    body2 = {"data": items[1:], "next": "{}?offset=2&limit=1".format(url)}
    responses.reset()
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body2),
        status=200,
        content_type="application/json",
        match_querystring=False,
    )

    assert next(it, None) == TrainingPredictionsRow(
        row_id=2,
        partition_id=2,
        prediction=0.22,
        prediction_values=[],
        timestamp=None,
        forecast_point=None,
        forecast_distance=None,
        series_id=None,
        prediction_explanations=None,
        shap_metadata=None,
    )
    assert len(responses.calls) == 1

    body = {"data": [], "next": "{}?offset=3&limit=1".format(url)}
    responses.reset()
    responses.add(
        responses.GET, url, body=json.dumps(body), status=200, content_type="application/json",
    )

    assert next(it, None) is None
    assert len(responses.calls) == 1


@responses.activate
@pytest.mark.parametrize(
    ["serializer", "response"],
    [
        pytest.param(
            "json",
            {
                "body": lambda url: json.dumps(
                    {
                        "data": [
                            {
                                "row_id": 1,
                                "partition_id": 2,
                                "prediction": 0.42,
                                "prediction_values": [],
                            },
                            {
                                "row_id": 2,
                                "partition_id": 2,
                                "prediction": 0.22,
                                "prediction_values": [],
                            },
                        ],
                        "next": "{}?limit=20".format(url),
                    }
                ),
                "content_type": "application/json",
            },
            id="json",
        ),
        pytest.param(
            "csv",
            {
                "body": lambda url: _write_csv(
                    [["row_id", "partition_id", "prediction"], [1, 2, 0.42], [2, 2, 0.22]]
                ),
                "content_type": "text/csv",
                "stream": True,
            },
            id="csv",
        ),
    ],
)
def test_get_all_as_dataframe__regression_two_predictions__ok(
    project_id, prediction_id, project_url, project_with_target_json, serializer, response,
):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    response["body"] = response["body"](url)
    responses.add(responses.GET, url, status=200, **response)
    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe(serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (2, 3)
    assert list(data_frame.columns) == [
        "row_id",
        "partition_id",
        "prediction",
    ]


@responses.activate
@pytest.mark.parametrize(
    ["serializer", "response"],
    [
        pytest.param(
            "json",
            {
                "body": lambda url: json.dumps(
                    {
                        "data": [
                            {
                                "row_id": 1,
                                "partition_id": "1960-01-01T00:00:00.000000Z",
                                "prediction": 0.42,
                                "prediction_values": [],
                                "timestamp": "1960-01-01T00:00:00.000000Z",
                                "forecast_point": "1959-12-01T00:00:00.000000Z",
                                "forecast_distance": 1,
                                "series_id": "1",
                            },
                            {
                                "row_id": 2,
                                "partition_id": "1960-01-01T00:00:00.000000Z",
                                "prediction": 0.22,
                                "prediction_values": [],
                                "timestamp": "1959-11-01T00:00:00.000000Z",
                                "forecast_point": "1959-12-01T00:00:00.000000Z",
                                "forecast_distance": 2,
                                "series_id": "1",
                            },
                        ],
                        "next": "{}?limit=20".format(url),
                    }
                ),
                "content_type": "application/json",
            },
            id="json",
        ),
        pytest.param(
            "csv",
            {
                "body": lambda url: _write_csv(
                    [
                        [
                            "row_id",
                            "partition_id",
                            "prediction",
                            "timestamp",
                            "forecast_point",
                            "forecast_distance",
                            "series_id",
                        ],
                        [
                            1,
                            "1960-01-01T00:00:00.000000Z",
                            0.42,
                            "1960-01-01T00:00:00.000000Z",
                            "1959-12-01T00:00:00.000000Z",
                            1,
                            "1",
                        ],
                        [
                            2,
                            "1960-01-01T00:00:00.000000Z",
                            0.22,
                            "1959-11-01T00:00:00.000000Z",
                            "1959-12-01T00:00:00.000000Z",
                            2,
                            "1",
                        ],
                    ]
                ),
                "content_type": "text/csv",
                "stream": True,
            },
            id="csv",
        ),
        pytest.param(
            "json",
            {
                "body": lambda url: json.dumps(
                    {
                        "data": [
                            {
                                "row_id": 1,
                                "partition_id": "1960-01-01T00:00:00.000000Z",
                                "prediction": 0.42,
                                "prediction_values": [],
                                "timestamp": "1960-01-01T00:00:00.000000Z",
                                "forecast_point": "1959-12-01T00:00:00.000000Z",
                                "forecast_distance": 1,
                                "series_id": 1,
                            },
                            {
                                "row_id": 2,
                                "partition_id": "1960-01-01T00:00:00.000000Z",
                                "prediction": 0.22,
                                "prediction_values": [],
                                "timestamp": "1959-11-01T00:00:00.000000Z",
                                "forecast_point": "1959-12-01T00:00:00.000000Z",
                                "forecast_distance": 2,
                                "series_id": 1,
                            },
                        ],
                        "next": "{}?limit=20".format(url),
                    }
                ),
                "content_type": "application/json",
            },
            id="json",
        ),
        pytest.param(
            "csv",
            {
                "body": lambda url: _write_csv(
                    [
                        [
                            "row_id",
                            "partition_id",
                            "prediction",
                            "timestamp",
                            "forecast_point",
                            "forecast_distance",
                            "series_id",
                        ],
                        [
                            1,
                            "1960-01-01T00:00:00.000000Z",
                            0.42,
                            "1960-01-01T00:00:00.000000Z",
                            "1959-12-01T00:00:00.000000Z",
                            1,
                            1,
                        ],
                        [
                            2,
                            "1960-01-01T00:00:00.000000Z",
                            0.22,
                            "1959-11-01T00:00:00.000000Z",
                            "1959-12-01T00:00:00.000000Z",
                            2,
                            1,
                        ],
                    ]
                ),
                "content_type": "text/csv",
                "stream": True,
            },
            id="csv",
        ),
    ],
)
def test_get_all_as_dataframe__timeseries_predictions__ok(
    project_id, prediction_id, project_url, project_with_target_json, serializer, response,
):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    response["body"] = response["body"](url)
    responses.add(responses.GET, url, status=200, **response)
    project_data = json.loads(project_with_target_json)
    project_data["partition"]["use_time_series"] = True
    responses.add(
        responses.GET,
        project_url,
        body=json.dumps(project_data),
        status=200,
        content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe(serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (2, 7)
    assert set(data_frame.columns) == {
        "row_id",
        "partition_id",
        "prediction",
        "timestamp",
        "forecast_point",
        "forecast_distance",
        "series_id",
    }

    assert len(data_frame) == 2
    assert None not in data_frame.timestamp.unique()
    assert None not in data_frame.forecast_point.unique()
    assert all(isinstance(value, int) for value in data_frame.forecast_distance)


@responses.activate
@pytest.mark.parametrize(
    ["serializer", "response"],
    [
        pytest.param(
            "json",
            {
                "body": lambda url: json.dumps(
                    {
                        "data": [
                            {
                                "row_id": 1,
                                "partition_id": 2,
                                "prediction": 0.42,
                                "prediction_values": [],
                            },
                            {
                                "row_id": 2,
                                "partition_id": 2,
                                "prediction": 0.22,
                                "prediction_values": [],
                            },
                        ],
                        "next": "{}?limit=20".format(url),
                    }
                ),
                "content_type": "application/json",
            },
            id="json",
        ),
        pytest.param(
            "csv",
            {
                "body": lambda url: _write_csv(
                    [["row_id", "partition_id", "prediction"], [1, 2, 0.42], [2, 2, 0.22]]
                ),
                "content_type": "text/csv",
                "stream": True,
            },
            id="csv",
        ),
    ],
)
def test_download_to_csv__two_regression_predictions__ok(
    project_id, prediction_id, project_url, project_with_target_json, serializer, response,
):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    response["body"] = response["body"](url)
    responses.add(responses.GET, url, status=200, **response)
    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    with tempfile.NamedTemporaryFile(mode="r") as a_file:
        obj.download_to_csv(a_file.name, serializer=serializer)
        data_frame = pd.read_csv(a_file.name)

    assert data_frame.shape == (2, 3)
    assert set(data_frame.columns) == {"row_id", "partition_id", "prediction"}


@responses.activate
def test_iterate_rows__one_classification_prediction__ok(project_id, prediction_id):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    items = [
        {
            "row_id": 1,
            "partition_id": 2,
            "prediction": 0.42,
            "prediction_values": [{"label": "good", "value": 0.8}, {"label": "bad", "value": 0.2}],
        },
    ]
    response = {"data": items, "next": "{}?offset=100&limit=100".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(response), status=200, content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)

    assert list(obj.iterate_rows()) == [
        TrainingPredictionsRow(
            row_id=1,
            partition_id=2,
            prediction=0.42,
            prediction_values=[{"label": "good", "value": 0.8}, {"label": "bad", "value": 0.2}],
            timestamp=None,
            forecast_point=None,
            forecast_distance=None,
            series_id=None,
            prediction_explanations=None,
            shap_metadata=None,
        ),
    ]
    assert len(responses.calls) == 1


@responses.activate
@pytest.mark.parametrize(
    ["serializer", "response"],
    [
        pytest.param(
            "json",
            {
                "body": lambda url: json.dumps(
                    {
                        "data": [
                            {
                                "row_id": 1,
                                "partition_id": 2,
                                "prediction": 0.42,
                                "prediction_values": [
                                    {"label": "good", "value": 0.8},
                                    {"label": "bad", "value": 0.2},
                                ],
                            }
                        ],
                        "next": "{}?limit=20".format(url),
                    }
                ),
                "content_type": "application/json",
            },
            id="json",
        ),
        pytest.param(
            "csv",
            {
                "body": lambda url: _write_csv(
                    [
                        ["row_id", "partition_id", "prediction", "foo_good", "foo_bad"],
                        [1, 2, 0.42, 0.8, 0.2],
                    ]
                ),
                "content_type": "text/csv",
                "stream": True,
            },
            id="csv",
        ),
    ],
)
def test_get_all_as_dataframe__one_classification_prediction__ok(
    project_id, prediction_id, project_url, project_with_target_json, serializer, response,
):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    response["body"] = response["body"](url)
    responses.add(responses.GET, url, status=200, **response)

    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe(class_prefix="foo_", serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (1, 5)
    assert set(data_frame.columns) == {
        "row_id",
        "partition_id",
        "prediction",
        "foo_good",
        "foo_bad",
    }


@responses.activate
@pytest.mark.parametrize(
    ["serializer", "response"],
    [
        pytest.param(
            "json",
            {
                "body": lambda url: json.dumps(
                    {
                        "data": [
                            {
                                "row_id": 1,
                                "partition_id": 2,
                                "prediction": 0.42,
                                "prediction_values": [
                                    {"label": "good", "value": 0.3},
                                    {"label": "bad", "value": 0.2},
                                    {"label": "ugly", "value": 0.2},
                                    {"label": "none", "value": 0.0},
                                ],
                            }
                        ],
                        "next": "{}?limit=20".format(url),
                    }
                ),
                "content_type": "application/json",
            },
            id="json",
        ),
        pytest.param(
            "csv",
            {
                "body": lambda url: _write_csv(
                    [
                        [
                            "row_id",
                            "partition_id",
                            "prediction",
                            "class_good",
                            "class_bad",
                            "class_ugly",
                            "class_none",
                        ],
                        [1, 2, 0.42, 0.3, 0.2, 0.2, 0.0],
                    ]
                ),
                "content_type": "text/csv",
                "stream": True,
            },
            id="csv",
        ),
    ],
)
def test_download_to_csv__multiclass_project__prediction__ok(
    project_id, prediction_id, project_url, project_with_target_json, serializer, response,
):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id,
    )
    response["body"] = response["body"](url)
    responses.add(responses.GET, url, status=200, **response)

    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    with tempfile.NamedTemporaryFile(mode="r") as a_file:
        obj.download_to_csv(a_file.name, serializer=serializer)
        data_frame = pd.read_csv(a_file.name)

    assert data_frame.shape == (1, 7)
    assert set(data_frame.columns) == {
        "row_id",
        "partition_id",
        "prediction",
        "class_good",
        "class_bad",
        "class_ugly",
        "class_none",
    }


@responses.activate
@pytest.mark.parametrize(
    ["serializer", "response"],
    [
        pytest.param(
            "json",
            {
                "body": lambda url: json.dumps(
                    {
                        "data": [
                            {
                                "row_id": 0,
                                "partition_id": "0.0",
                                "prediction": 0.0,
                                "prediction_threshold": 0.5,
                                "prediction_values": [
                                    {"label": 1.0, "value": 0.311178247734139},
                                    {"label": 0.0, "value": 0.6888217522658611},
                                ],
                                "prediction_explanations": [
                                    {
                                        "featureValue": "5",
                                        "strength": -0.07643818895488697,
                                        "feature": "number_diagnoses",
                                        "label": 1.0,
                                    },
                                ],
                                "shap_metadata": {
                                    "shap_base_value": 0.39640626311302185,
                                    "shap_remaining_total": 0.06644934718939229,
                                    "warnings": None,
                                },
                            }
                        ],
                        "next": "{}?limit=20".format(url),
                    }
                ),
                "content_type": "application/json",
            },
            id="json",
        ),
        pytest.param(
            "csv",
            {
                "body": lambda url: _write_csv(
                    [
                        [
                            "row_id",
                            "partition_id",
                            "prediction",
                            "class_1.0",
                            "class_0.0",
                            "explained_class",
                            "Explanation_1_feature_name",
                            "Explanation_1_feature_value",
                            "Explanation_1_strength",
                            "shap_remaining_total",
                            "shap_base_value",
                        ],
                        [
                            0,
                            0.0,
                            0.0,
                            0.311178247734139,
                            0.6888217522658611,
                            1.0,
                            "number_diagnoses",
                            5,
                            -0.07643818895488697,
                            "number_inpatient",
                            0,
                            -0.05988679910481005,
                            0.06644934718939229,
                            0.396406263113,
                        ],
                    ]
                ),
                "content_type": "text/csv",
                "stream": True,
            },
            id="csv",
        ),
    ],
)
def test_get_all_as_dataframe_one_classification_prediction_with_explanations(
    project_id, prediction_id, project_url, project_with_target_json, serializer, response
):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id
    )
    response["body"] = response["body"](url)
    responses.add(responses.GET, url, status=200, **response)

    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type="application/json",
    )

    obj = TrainingPredictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe(serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (1, 11)
    assert set(data_frame.columns) == {
        "row_id",
        "partition_id",
        "prediction",
        "class_1.0",
        "class_0.0",
        "explained_class",
        "shap_remaining_total",
        "shap_base_value",
        "Explanation_1_feature_name",
        "Explanation_1_feature_value",
        "Explanation_1_strength",
    }


@responses.activate
@pytest.mark.parametrize(
    ["serializer", "response"],
    [
        pytest.param(
            "json",
            {
                "body": lambda url: json.dumps(
                    {
                        "data": [
                            {
                                "row_id": 0,
                                "partition_id": "0.0",
                                "prediction": 71700.0,
                                "prediction_values": [{"value": 71700.0, "label": "SalePrice"}],
                                "shap_metadata": {
                                    "shap_base_value": 31108.28125,
                                    "shap_remaining_total": -4045.506156357904,
                                    "warnings": None,
                                },
                                "prediction_explanations": [
                                    {
                                        "feature_value": "172632",
                                        "strength": 9021.030323424387,
                                        "feature": "MachineID",
                                        "label": "SalePrice",
                                    },
                                    {
                                        "feature_value": "3855",
                                        "strength": -4883.805345350174,
                                        "feature": "ModelID",
                                        "label": "SalePrice",
                                    },
                                ],
                            },
                            {
                                "row_id": 1,
                                "partition_id": "0.0",
                                "prediction": 12275.0,
                                "prediction_values": [{"value": 12275.0, "label": "SalePrice"}],
                                "shap_metadata": {
                                    "shap_base_value": 31108.28125,
                                    "shap_remaining_total": -3350.9689739794676,
                                    "warnings": None,
                                },
                                "prediction_explanations": [
                                    {
                                        "feature_value": "1986",
                                        "strength": -6307.912477762799,
                                        "feature": "YearMade",
                                        "label": "SalePrice",
                                    },
                                    {
                                        "feature_value": "1190700",
                                        "strength": -3257.09113279142,
                                        "feature": "MachineID",
                                        "label": "SalePrice",
                                    },
                                ],
                            },
                        ],
                        "next": "{}?limit=20".format(url),
                    }
                ),
                "content_type": "application/json",
            },
            id="json",
        ),
        pytest.param(
            "csv",
            {
                "body": lambda url: _write_csv(
                    [
                        [
                            "row_id",
                            "partition_id",
                            "prediction",
                            "Explanation_1_feature_name",
                            "Explanation_1_feature_value",
                            "Explanation_1_strength",
                            "Explanation_2_feature_name",
                            "Explanation_2_feature_value",
                            "Explanation_2_strength",
                            "shap_remaining_total",
                            "shap_base_value",
                        ],
                        [
                            0,
                            0.0,
                            71700.0,
                            "MachineID",
                            "172632",
                            9021.030323424387,
                            "ModelID",
                            "3855",
                            -4883.805345350174,
                            -4045.506156357904,
                            31108.28125,
                        ],
                        [
                            1,
                            0.0,
                            12275.0,
                            "YearMade",
                            "1986",
                            -6307.912477762799,
                            "MachineID",
                            "1190700",
                            -3257.09113279142,
                            -3350.9689739794676,
                            31108.28125,
                        ],
                    ]
                ),
                "content_type": "text/csv",
                "stream": True,
            },
            id="csv",
        ),
    ],
)
def test_get_all_as_dataframe_regression_predictions_with_explanations(
    project_id, prediction_id, project_url, project_with_target_json, serializer, response
):
    url = "https://host_name.com/projects/{}/trainingPredictions/{}/".format(
        project_id, prediction_id
    )
    response["body"] = response["body"](url)
    responses.add(responses.GET, url, status=200, **response)
    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type="application/json",
    )

    predictions = TrainingPredictions.get(project_id, prediction_id)
    data_frame = predictions.get_all_as_dataframe(serializer=serializer)

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (2, 11)
    assert list(data_frame.columns) == [
        "row_id",
        "partition_id",
        "prediction",
        "Explanation_1_feature_name",
        "Explanation_1_feature_value",
        "Explanation_1_strength",
        "Explanation_2_feature_name",
        "Explanation_2_feature_value",
        "Explanation_2_strength",
        "shap_remaining_total",
        "shap_base_value",
    ]



# coding: utf-8

import numpy as np
import pandas as pd
import pytest
import responses

from datarobot import Feature
from datarobot.models.pairwise_statistics import (
    PairwiseConditionalProbabilities,
    PairwiseCorrelations,
    PairwiseJointProbabilities,
    PairwiseProbabilitiesBase,
)
from datarobot.utils import from_api
from tests.test_features import get_feature_response_multicategorical


class TestPairwiseStatistics(object):
    @pytest.fixture()
    def correlation_response(self):
        # Wrong lexicographical order to check if output gets ordered
        label_names = [u"🐶", "A"]
        statistic_values = np.array([[np.nan, 1.0], [2.0, 3.0]])
        data = self._get_data_field_for_statistics_api_response(
            label_names, statistic_values, label_relevance=None
        )
        response = {
            "statisticType": "correlation",
            "projectId": "5f8461253b8f1a2bd46ab5aa",
            "featureName": "multicategorical_column",
            "data": data,
        }
        return response

    @pytest.fixture()
    def joint_probability_response(self):
        data = self._get_data_field_for_probabilities_response()
        response = {
            "statisticType": "jointProbability",
            "projectId": "5f8461253b8f1a2bd46ab5aa",
            "featureName": "multicategorical_column",
            "data": data,
        }
        return response

    @pytest.fixture()
    def conditional_probability_response(self):
        data = self._get_data_field_for_probabilities_response()
        response = {
            "statisticType": "conditionalProbability",
            "projectId": "5f8461253b8f1a2bd46ab5aa",
            "featureName": "multicategorical_column",
            "data": data,
        }
        return response

    @pytest.fixture()
    def pairwise_correlation_instance(self, correlation_response):
        converted_response = from_api(correlation_response)
        return PairwiseCorrelations(
            feature_name=converted_response["feature_name"], data=converted_response["data"]
        )

    @pytest.fixture()
    def pairwise_joint_probabilities_instance(self, joint_probability_response):
        converted_response = from_api(joint_probability_response)
        return PairwiseJointProbabilities(
            feature_name=converted_response["feature_name"], data=converted_response["data"]
        )

    @pytest.fixture()
    def pairwise_conditional_probabilities_instance(self, conditional_probability_response):
        converted_response = from_api(conditional_probability_response)
        return PairwiseConditionalProbabilities(
            feature_name=converted_response["feature_name"], data=converted_response["data"]
        )

    def _get_data_field_for_probabilities_response(self):
        data = []
        for multiplier, relevance_configuration in enumerate(
            [(0, 0), (0, 1), (1, 0), (1, 1)], start=1
        ):
            # Wrong lexicographical order to check if output gets ordered
            label_names = [u"🐶", "A"]
            statistic_values = np.array([[np.nan, 1.0], [2.0, 3.0]])
            data.extend(
                self._get_data_field_for_statistics_api_response(
                    label_names, statistic_values * multiplier, relevance_configuration
                )
            )
        return data

    def _get_data_field_for_statistics_api_response(
        self, label_names, statistic_values, label_relevance=None
    ):
        """
        Creates the content of a pairwise statistics response ``data`` field for one relevance
        configuration.

        Parameters
        ----------
        label_names : list(str)
        statistic_values : numpy.ndarray
        label_relevance : tuple(int), optional
            Label relevance. The value of label_relevance[0] gets assigned to the first label of
            the label_configuration, label_relevance[1] gets assigned to the second label. By
            default no relevance gets assigned (required for correlation).

        Returns
        -------
        list(dict)
            Content of the pairwise statistics response ``data`` field for one relevance
            configuration, following the schema of the API response.
        """
        data = []
        for label_name_column, statistic_column in zip(label_names, statistic_values):
            for label_name_row, statistic_value in zip(label_names, statistic_column):
                if label_relevance:
                    label_configuration = [
                        {"label": label_name_column, "relevance": label_relevance[0]},
                        {"label": label_name_row, "relevance": label_relevance[1]},
                    ]
                else:
                    label_configuration = [{"label": label_name_column}, {"label": label_name_row}]
                if statistic_value is None or np.isnan(statistic_value):
                    # If the statistic_value is missing, there will be no statisticValue key
                    data.append({"labelConfiguration": label_configuration})
                else:
                    data.append(
                        {
                            "statisticValue": statistic_value,
                            "labelConfiguration": label_configuration,
                        }
                    )
        return data

    def test_to_statistic_dataframe(self, correlation_response):
        data = from_api(correlation_response["data"])
        actual = PairwiseCorrelations._to_statistic_dataframe(data)

        self._assert_statistic_dataframe(actual)

    def _assert_statistic_dataframe(self, actual):
        values = np.array([[3.0, 2.0], [1.0, np.nan]])
        expected = pd.DataFrame(values, index=[u"A", u"🐶"], columns=[u"A", u"🐶"])
        pd.testing.assert_frame_equal(actual, expected)

    def test_to_statistic_dataframes(self, joint_probability_response):
        data = from_api(joint_probability_response["data"])
        actual = PairwiseProbabilitiesBase._to_statistic_dataframes(data)

        self._assert_statistic_dataframes(actual)

    def _assert_statistic_dataframes(self, actual):
        expected = {}
        # Indexes and columns are expected to be sorted lexicographical
        values = np.array([[3.0, 2.0], [1.0, np.nan]])
        sorted_labels = [u"A", u"🐶"]
        expected[(0, 0)] = pd.DataFrame(values, index=sorted_labels, columns=sorted_labels)
        expected[(0, 1)] = pd.DataFrame(values * 2, index=sorted_labels, columns=sorted_labels)
        expected[(1, 0)] = pd.DataFrame(values * 3, index=sorted_labels, columns=sorted_labels)
        expected[(1, 1)] = pd.DataFrame(values * 4, index=sorted_labels, columns=sorted_labels)
        assert len(actual) == len(expected)
        for label_configuration in expected.keys():
            pd.testing.assert_frame_equal(
                actual[label_configuration], expected[label_configuration]
            )

    @responses.activate
    def test_get_correlations(self, correlation_response):
        multilabel_insights_key = "5f8461253b8f1a2bd46ab5aa"
        responses.add(
            responses.GET,
            (
                "https://host_name.com/multilabelInsights/"
                "{}/pairwiseStatistics/?statisticType=correlation"
            ).format(multilabel_insights_key),
            json=correlation_response,
            match_querystring=True,
        )
        pairwise_correlations = PairwiseCorrelations.get(multilabel_insights_key)

        self._assert_correlations(pairwise_correlations, correlation_response)

    def _assert_correlations(self, pairwise_correlations, correlation_response):
        assert pairwise_correlations.values == from_api(correlation_response["data"])
        assert (
            str(pairwise_correlations)
            == "PairwiseCorrelations(feature_name=multicategorical_column)"
        )
        self._assert_statistic_dataframe(pairwise_correlations.statistic_dataframe)

    @responses.activate
    def test_get_correlations_via_feature(self, correlation_response):
        multilabel_insights_key = "5f8461253b8f1a2bd46ab5aa"
        project_id = "5f8461253b8f1a2bd46ab5ab"
        feature_name = "multicategorical_column"
        feature_response_multicategorical = get_feature_response_multicategorical(
            multilabel_insights_key, project_id, feature_name
        )
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/features/{}/".format(project_id, feature_name),
            json=feature_response_multicategorical,
        )
        responses.add(
            responses.GET,
            (
                "https://host_name.com/multilabelInsights/"
                "{}/pairwiseStatistics/?statisticType=correlation"
            ).format(multilabel_insights_key),
            json=correlation_response,
            match_querystring=True,
        )
        feature = Feature.get(project_id, feature_name)
        pairwise_correlations = feature.get_pairwise_correlations()
        self._assert_correlations(pairwise_correlations, correlation_response)

    def test_correlation_as_dataframe(self, pairwise_correlation_instance):
        expected = pairwise_correlation_instance.statistic_dataframe
        actual = pairwise_correlation_instance.as_dataframe()

        pd.testing.assert_frame_equal(actual, expected)

    @responses.activate
    def test_get_joint_probabilities(self, joint_probability_response):
        multilabel_insights_key = "5f8461253b8f1a2bd46ab5aa"
        responses.add(
            responses.GET,
            (
                "https://host_name.com/multilabelInsights/"
                "{}/pairwiseStatistics/?statisticType=jointProbability"
            ).format(multilabel_insights_key),
            json=joint_probability_response,
            match_querystring=True,
        )
        pairwise_joint_probabilities = PairwiseJointProbabilities.get(multilabel_insights_key)

        self._assert_joint_probabilities(pairwise_joint_probabilities, joint_probability_response)

    def _assert_joint_probabilities(self, pairwise_joint_probabilities, joint_probability_response):
        assert pairwise_joint_probabilities.values == from_api(joint_probability_response["data"])
        assert (
            str(pairwise_joint_probabilities)
            == "PairwiseJointProbabilities(feature_name=multicategorical_column)"
        )
        self._assert_statistic_dataframes(pairwise_joint_probabilities.statistic_dataframes)

    @responses.activate
    def test_get_joint_probabilities_via_feature(self, joint_probability_response):
        multilabel_insights_key = "5f8461253b8f1a2bd46ab5aa"
        project_id = "5f8461253b8f1a2bd46ab5ab"
        feature_name = "multicategorical_column"
        feature_response_multicategorical = get_feature_response_multicategorical(
            multilabel_insights_key, project_id, feature_name
        )
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/features/{}/".format(project_id, feature_name),
            json=feature_response_multicategorical,
        )
        responses.add(
            responses.GET,
            (
                "https://host_name.com/multilabelInsights/"
                "{}/pairwiseStatistics/?statisticType=jointProbability"
            ).format(multilabel_insights_key),
            json=joint_probability_response,
            match_querystring=True,
        )
        feature = Feature.get(project_id, feature_name)
        pairwise_joint_probabilities = feature.get_pairwise_joint_probabilities()
        self._assert_joint_probabilities(pairwise_joint_probabilities, joint_probability_response)

    @responses.activate
    def test_get_conditional_probabilities(self, conditional_probability_response):
        multilabel_insights_key = "5f8461253b8f1a2bd46ab5aa"
        responses.add(
            responses.GET,
            (
                "https://host_name.com/multilabelInsights/"
                "{}/pairwiseStatistics/?statisticType=conditionalProbability"
            ).format(multilabel_insights_key),
            json=conditional_probability_response,
            match_querystring=True,
        )
        pairwise_conditional_probabilities = PairwiseConditionalProbabilities.get(
            multilabel_insights_key
        )

        self._assert_conditional_probabilities(
            pairwise_conditional_probabilities, conditional_probability_response
        )

    def _assert_conditional_probabilities(
        self, pairwise_conditional_probabilities, conditional_probability_response
    ):
        assert pairwise_conditional_probabilities.values == from_api(
            conditional_probability_response["data"]
        )
        assert (
            str(pairwise_conditional_probabilities)
            == "PairwiseConditionalProbabilities(feature_name=multicategorical_column)"
        )
        self._assert_statistic_dataframes(pairwise_conditional_probabilities.statistic_dataframes)

    @responses.activate
    def test_get_conditional_probabilities_via_feature(self, conditional_probability_response):
        multilabel_insights_key = "5f8461253b8f1a2bd46ab5aa"
        project_id = "5f8461253b8f1a2bd46ab5ab"
        feature_name = "multicategorical_column"
        feature_response_multicategorical = get_feature_response_multicategorical(
            multilabel_insights_key, project_id, feature_name
        )
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/features/{}/".format(project_id, feature_name),
            json=feature_response_multicategorical,
        )
        responses.add(
            responses.GET,
            (
                "https://host_name.com/multilabelInsights/"
                "{}/pairwiseStatistics/?statisticType=conditionalProbability"
            ).format(multilabel_insights_key),
            json=conditional_probability_response,
            match_querystring=True,
        )
        feature = Feature.get(project_id, feature_name)
        pairwise_joint_probabilities = feature.get_pairwise_conditional_probabilities()
        self._assert_conditional_probabilities(
            pairwise_joint_probabilities, conditional_probability_response
        )

    @pytest.mark.parametrize(
        "fixture_name",
        ["pairwise_conditional_probabilities_instance", "pairwise_joint_probabilities_instance"],
    )
    def test_pairwise_probabilities_as_dataframe(self, request, fixture_name):
        for label_configuration in [(0, 0), (0, 1), (1, 0), (1, 1)]:
            pairwise_statistics_instance = request.getfixturevalue(fixture_name)
            actual = pairwise_statistics_instance.as_dataframe(label_configuration)
            expected = pairwise_statistics_instance.statistic_dataframes[label_configuration]
            pd.testing.assert_frame_equal(expected, actual)

    @pytest.mark.parametrize(
        "fixture_name",
        ["pairwise_conditional_probabilities_instance", "pairwise_joint_probabilities_instance"],
    )
    def test_pairwise_probabilities_as_dataframe_not_a_valid_label_configuration(
        self, request, fixture_name
    ):
        pairwise_statistics_instance = request.getfixturevalue(fixture_name)
        invalid_label_configuration = (2, -1)
        expected_error_msg = (
            "You have passed an invalid label configuration. Valid options are "
            "\(0, 0\), \(0, 1\), \(1, 0\) and \(1, 1\)"
        )
        with pytest.raises(ValueError, match=expected_error_msg):
            pairwise_statistics_instance._as_dataframe(invalid_label_configuration)



import json

import responses
from six.moves.urllib.parse import urljoin

from datarobot.models.visualai import DuplicateImage, Image, SampleImage


@responses.activate
def test_get_image(visualai_image_url, visualai_image_id, visualai_image, project_id):
    responses.add(
        responses.GET,
        visualai_image_url,
        status=200,
        content_type="application/json",
        body=json.dumps(visualai_image),
    )
    image = Image.get(project_id, visualai_image_id)
    assert isinstance(image, Image)
    assert image.id == visualai_image["imageId"]
    assert image.project_id == project_id
    assert image.width == visualai_image["width"]
    assert image.height == visualai_image["height"]


@responses.activate
def test_get_image_bytes(visualai_image_url, visualai_image_id, visualai_image_file, project_id):
    responses.add(
        responses.GET,
        urljoin(visualai_image_url, "file/"),
        status=200,
        content_type="image/jpeg",
        body=visualai_image_file,
    )
    image = Image(project_id=project_id, image_id=visualai_image_id, width=256, height=256)
    assert image.image_type == "image/jpeg"
    assert image.image_bytes == visualai_image_file


@responses.activate
def test_get_sample_images(visualai_project, visualai_sample, project_url, project_id):
    responses.add(
        responses.GET,
        project_url,
        status=200,
        content_type="application/json",
        body=json.dumps(visualai_project),
    )
    stage_url = {
        "aim": "imageSamples/?featureName=image",
        "eda": "imageSamples/?featureName=image",
        "eda2": "images/?featureName=image",
        "empty": "imageSamples/?featureName=image",
        "modeling": "images/?featureName=image",
    }[visualai_project["stage"]]
    responses.add(
        responses.GET,
        urljoin(project_url, stage_url),
        status=200,
        content_type="application/json",
        body=json.dumps(visualai_sample),
    )
    img0, img1 = SampleImage.list(project_id, "image")
    assert isinstance(img0, SampleImage)
    assert img0.image.id == visualai_sample.get("data")[0].get("imageId")
    assert img0.project_id == project_id
    assert img0.target_value == visualai_sample.get("data")[0].get("targetValue")
    assert isinstance(img1, SampleImage)
    assert img1.image.id == visualai_sample.get("data")[1].get("imageId")
    assert img1.project_id == project_id
    assert img1.target_value == visualai_sample.get("data")[1].get("targetValue")


@responses.activate
def test_get_duplicate_images(visualai_project, visualai_duplicate, project_url, project_id):
    responses.add(
        responses.GET,
        project_url,
        status=200,
        content_type="application/json",
        body=json.dumps(visualai_project),
    )
    responses.add(
        responses.GET,
        urljoin(project_url, "duplicateImages/image/"),
        status=200,
        content_type="application/json",
        body=json.dumps(visualai_duplicate),
    )
    img0, img1 = DuplicateImage.list(project_id, "image")
    assert isinstance(img0, DuplicateImage)
    assert img0.image.id == visualai_duplicate.get("data")[0].get("imageId")
    assert img0.project_id == project_id
    assert img0.count == visualai_duplicate.get("data")[0].get("rowCount")
    assert isinstance(img1, DuplicateImage)
    assert img1.image.id == visualai_duplicate.get("data")[1].get("imageId")
    assert img1.project_id == project_id
    assert img1.count == visualai_duplicate.get("data")[1].get("rowCount")



import json

import pytest
import responses


@pytest.fixture()
def deployments_data():
    deployments = [
        {
            "id": "5c6f7a9e9ca9b20017ff95a2",
            "label": "is_bad Predictions",
            "description": "",
            "model": {
                "projectId": "5c0a969859b00004ba52e41b",
                "targetName": "is_bad",
                "id": "5cb9c8b979872d014a0bb080",
                "type": "TensorFlow Deep Learning Classifier",
            },
            "defaultPredictionServer": {
                "id": "5a7c7a72192765001fc22769",
                "url": "https://pred.example.com",
                "datarobot-key": "27782",
            },
            "predictionUsage": {
                "dailyRates": [5, 2, 10, 5, 13, 4, 7],
                "lastTimestamp": "2019-04-22T04:03:45.000000Z",
            },
            "capabilities": {"supportsModelReplacement": True, "supportsDriftTracking": True},
            "permissions": [
                "CAN_SHARE",
                "CAN_VIEW",
                "CAN_DELETE_DEPLOYMENT",
                "CAN_MAKE_PREDICTIONS",
            ],
            "serviceHealth": {"status": "unknown"},
            "modelHealth": {
                "status": "unknown",
                "message": "1 important feature is drifting",
                "startDate": "2019-04-18T02:00:00.000000Z",
                "endDate": "2019-04-22T02:00:00.000000Z",
            },
            "accuracyHealth": {"status": "unavailable", "startDate": None, "endDate": None},
        },
        {
            "id": "5c6f7a9e9ca9b20017ff93b7",
            "label": "is_bad Predictions2",
            "description": None,
            "model": {
                "projectId": "5c0a969859b00004ba52e41b",
                "targetName": "is_bad",
                "id": "5cb9c8b979872d014a0bb080",
                "type": "TensorFlow Deep Learning Classifier",
            },
            "defaultPredictionServer": {
                "id": "5a7c7a72192765001fc22769",
                "url": "https://pred.example.com",
                "datarobot-key": "27782",
            },
            "predictionUsage": {"dailyRates": [], "lastTimestamp": None},
            "capabilities": {"supportsModelReplacement": False, "supportsDriftTracking": False},
            "permissions": [
                "CAN_SHARE",
                "CAN_VIEW",
                "CAN_DELETE_DEPLOYMENT",
                "CAN_MAKE_PREDICTIONS",
            ],
            "serviceHealth": {"status": "unknown"},
            "modelHealth": {"status": "unknown", "startDate": None, "endDate": None},
            "accuracyHealth": {"status": "unavailable", "startDate": None, "endDate": None},
        },
    ]
    return deployments


@pytest.fixture()
def deployment_data(deployments_data):
    return deployments_data[0]


@pytest.fixture()
def deployment_get_response(unittest_endpoint, deployments_data):
    for deployment_data in deployments_data:
        url = "{}/{}/{}/".format(unittest_endpoint, "deployments", deployment_data["id"])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(deployment_data),
        )



import json

import pytest
import responses

from datarobot import errors, FeatureAssociationMatrixDetails


@pytest.fixture
def feature1(feature_association_matrix_details_data):
    return feature_association_matrix_details_data["features"][0]


@pytest.fixture
def feature2(feature_association_matrix_details_data):
    return feature_association_matrix_details_data["features"][1]


@responses.activate
def test_get_feature_association_details(
    project_id,
    feature1,
    feature2,
    feature_association_matrix_details_url,
    feature_association_matrix_details_data,
):
    fam_url = feature_association_matrix_details_url.format(project_id, feature1, feature2)
    responses.add(
        responses.GET,
        fam_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_matrix_details_data),
    )
    fam_details = FeatureAssociationMatrixDetails.get(
        project_id=project_id, feature1=feature1, feature2=feature2
    )

    assert set(fam_details.to_dict().keys()) == {"chart_type", "values", "features", "types"}
    assert len(fam_details.values) == len(feature_association_matrix_details_data["values"])
    assert fam_details.features == feature_association_matrix_details_data["features"]
    assert fam_details.types == feature_association_matrix_details_data["types"]
    assert fam_details.chart_type == feature_association_matrix_details_data["chart_type"]


@responses.activate
def test_get_feature_association_details__project_not_found(
    feature1, feature2, feature_association_matrix_details_url,
):
    non_existing_pid = "6034fcd7b9bc24be51493ad0"
    error = {"message": "Project with ID {} doesn't exist".format(non_existing_pid)}
    fam_url = feature_association_matrix_details_url.format(non_existing_pid, feature1, feature2)
    responses.add(
        responses.GET, fam_url, status=404, content_type="application/json", body=json.dumps(error),
    )
    with pytest.raises(
        errors.ClientError, match="Project with ID {} doesn't exist".format(non_existing_pid)
    ):
        FeatureAssociationMatrixDetails.get(
            project_id=non_existing_pid, feature1=feature1, feature2=feature2
        )


@responses.activate
def test_get_feature_association_details__feature_not_found(
    project_id, feature1, feature_association_matrix_details_url,
):
    non_existing_feature = "fake"
    error = {"message": "Invalid features were passed."}
    fam_url = feature_association_matrix_details_url.format(
        project_id, non_existing_feature, feature2
    )
    responses.add(
        responses.GET, fam_url, status=404, content_type="application/json", body=json.dumps(error),
    )
    with pytest.raises(errors.ClientError, match="Invalid features were passed."):
        FeatureAssociationMatrixDetails.get(
            project_id=project_id, feature1=non_existing_feature, feature2=feature2
        )



from copy import deepcopy
import json
import os
import tempfile

import pytest
import responses

from datarobot import TARGET_TYPE
from datarobot._experimental import CustomTrainingBlueprint


@pytest.fixture
def mocked_training_model(mocked_models):
    return mocked_models["data"][0]


@pytest.fixture
def mocked_custom_model_version():
    return {
        "id": "5cf4d3f5f930e26daac18aBB",
        "customModelId": "5cf4d3f5f930e26daac18a1BB",
        "label": "Version 2",
        "description": "ss",
        "versionMinor": 2,
        "versionMajor": 3,
        "isFrozen": True,
        "items": [],
        "created": "2019-09-28T15:19:26.587583Z",
    }


@pytest.fixture
def mocked_environment():
    return {
        "id": "5cf4d3f5f930e26daac18a1a",
        "name": "env1",
        "description": "some description",
        "programmingLanguage": "python",
        "isPublic": True,
        "created": "2019-09-28T15:19:26.587583Z",
        "latestVersion": {
            "id": "5cf4d3f5f930e26daac18a1a",
            "environmentId": "5cf4d3f5f930e26daac18a1xx",
            "label": "Version 1",
            "description": "",
            "buildStatus": "success",
            "created": "2019-09-28T15:19:26.587583Z",
        },
    }


@pytest.fixture
def mocked_blueprints(mocked_training_model, mocked_environment):
    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "previous": None,
        "data": [
            {
                "userBlueprintId": "5cf4d3f5f930e26daac18a1a",
                "customModel": {"id": mocked_training_model["id"], "name": "cm"},
                "customModelVersion": {"id": "5cf4d3f5f930e26daac18acv", "label": "cmv"},
                "executionEnvironment": {"id": mocked_environment["id"], "name": "ee"},
                "executionEnvironmentVersion": {"id": "5cf4d3f5f930e26daac18aev", "label": "eev"},
                "training_history": [],
            },
            {
                "userBlueprintId": "5cf4d3f5f930e26daac18a2b",
                "customModel": {"id": "5cf4d3f5f930e26daac18bcc", "name": "cm n2"},
                "customModelVersion": {"id": "5cf4d3f5f930e26daac18bcv", "label": "cmv n2"},
                "executionEnvironment": {"id": "5cf4d3f5f930e26daac18bee", "name": "ee n2"},
                "executionEnvironmentVersion": {
                    "id": "5cf4d3f5f930e26daac18bev",
                    "label": "eev n2",
                },
                "training_history": [
                    {
                        "pid": "5cf4d3f5f930e26daac18abc",
                        "project_name": "my_project",
                        "lid": "5cf4d3f5f930e26daac18bbbe",
                        "creation_date": "2019-09-07T15:50-04:00",
                        "project_models_count": 3,
                        "target_name": "my_target",
                    },
                    {
                        "pid": "5cf4d3f5f930e26daac18abc",
                        "project_name": "my_project",
                        "lid": "5cf4d3f5f930e26daac18cce",
                        "creation_date": "2019-09-07T15:50-06:00",
                        "project_models_count": 3,
                        "target_name": "my_target",
                    },
                ],
            },
        ],
    }


@pytest.fixture
def mocked_blueprint(mocked_blueprints):
    return mocked_blueprints["data"][0]


@pytest.fixture
def make_blueprints_url(unittest_endpoint):
    def _make_blueprints_url(image_id=None):
        base_url = "{}/customTrainingBlueprints/".format(unittest_endpoint)
        if image_id is not None:
            return "{}{}/".format(base_url, image_id)
        return base_url

    return _make_blueprints_url


def assert_blueprint(blueprint, blueprint_json):
    assert blueprint.id == blueprint_json["userBlueprintId"]

    cm_ver = blueprint_json["customModelVersion"]
    assert blueprint.custom_model_version["id"] == cm_ver["id"]
    assert blueprint.custom_model_version["label"] == cm_ver["label"]


def mock_get_response(url, response):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(response),
    )


def test_from_server_data(mocked_blueprint):
    blueprint = CustomTrainingBlueprint.from_server_data(mocked_blueprint)
    assert_blueprint(blueprint, mocked_blueprint)


@responses.activate
def test_create_blueprint(mocked_blueprint, make_blueprints_url):
    responses.add(
        responses.POST,
        make_blueprints_url(),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_blueprint),
    )

    kwargs = dict(custom_model_version_id=mocked_blueprint["customModelVersion"]["id"],)
    blueprint = CustomTrainingBlueprint.create(**kwargs)
    assert_blueprint(blueprint, mocked_blueprint)

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_blueprints_url()


@responses.activate
def test_create_blueprint_use_latest_version(
    mocked_blueprint, make_blueprints_url, mocked_training_model, mocked_environment
):

    responses.add(
        responses.GET,
        "https://host_name.com/customModels/{}/".format(mocked_training_model["id"]),
        body=json.dumps(mocked_training_model),
        status=200,
        content_type="application/json",
    )

    responses.add(
        responses.POST,
        make_blueprints_url(),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_blueprint),
    )

    CustomTrainingBlueprint.create(custom_model_id=mocked_blueprint["customModel"]["id"])

    assert responses.calls[0].request.method == "GET"
    assert "customModels" in responses.calls[0].request.url
    assert responses.calls[1].request.method == "POST"
    assert responses.calls[1].request.url == make_blueprints_url()
    assert (
        json.loads(responses.calls[1].request.body)["customModelVersionId"]
        == mocked_training_model["latestVersion"]["id"]
    )


@responses.activate
def test_create_clean(
    mocked_environment,
    mocked_training_model,
    mocked_custom_model_version,
    mocked_blueprint,
    make_blueprints_url,
):
    envdir = tempfile.mkdtemp()
    codedir = os.path.abspath(tempfile.mkdtemp())
    bfile = os.path.join(codedir, "bfile")
    cfile = os.path.join(codedir, "cfile")
    open(os.path.join(envdir, "afile"), "w").write("content")
    open(bfile, "w").write("content")
    open(cfile, "w").write("content")

    responses.add(
        responses.POST,
        "https://host_name.com/executionEnvironments/",
        body=json.dumps(mocked_environment),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        "https://host_name.com/executionEnvironments/{}/".format(mocked_environment["id"]),
        body=json.dumps(mocked_environment),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.POST,
        "https://host_name.com/executionEnvironments/{}/versions/".format(mocked_environment["id"]),
        body=json.dumps({"id": "abc123"}),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        "https://host_name.com/executionEnvironments/{}/versions/{}/".format(
            mocked_environment["id"], "abc123"
        ),
        body=json.dumps(
            {"build_status": "success", "environment_id": mocked_environment["id"], "id": "abc123"}
        ),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.POST,
        "https://host_name.com/customModels/",
        body=json.dumps(mocked_training_model),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.POST,
        "https://host_name.com/customModels/{}/versions/".format(mocked_training_model["id"]),
        body=json.dumps(mocked_custom_model_version),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.POST,
        make_blueprints_url(),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_blueprint),
    )
    blueprint = CustomTrainingBlueprint.create_from_scratch(
        name="hi",
        environment_dir=envdir,
        training_code_files=[bfile, cfile],
        target_type=TARGET_TYPE.REGRESSION,
    )
    assert_blueprint(blueprint, mocked_blueprint)


@responses.activate
def test_list_blueprints(mocked_blueprints, make_blueprints_url):
    url = make_blueprints_url()
    mock_get_response(url, mocked_blueprints)

    blueprints = CustomTrainingBlueprint.list()

    assert len(blueprints) == len(mocked_blueprints["data"])
    for image, mocked_image in zip(blueprints, mocked_blueprints["data"]):
        assert_blueprint(image, mocked_image)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_blueprints_multiple_pages(mocked_blueprints, make_blueprints_url):
    url1 = make_blueprints_url()
    url2 = make_blueprints_url() + "2"

    mocked_images_2nd = deepcopy(mocked_blueprints)
    mocked_blueprints["next"] = url2

    mock_get_response(url1, mocked_blueprints)
    mock_get_response(url2, mocked_images_2nd)

    blueprints = CustomTrainingBlueprint.list()
    assert len(blueprints) == len(mocked_blueprints["data"]) + len(mocked_images_2nd["data"])
    for image, mocked_image in zip(
        blueprints, mocked_blueprints["data"] + mocked_images_2nd["data"]
    ):
        assert_blueprint(image, mocked_image)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith(url1)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(url2)


@responses.activate
def test_from_dropin(
    mocked_environment,
    mocked_custom_model_version,
    make_blueprints_url,
    mocked_training_model,
    mocked_blueprint,
):
    envdir = tempfile.mkdtemp()
    codedir = os.path.abspath(tempfile.mkdtemp())
    bfile = os.path.join(codedir, "bfile")
    cfile = os.path.join(codedir, "cfile")
    open(os.path.join(envdir, "afile"), "w").write("content")
    open(bfile, "w").write("content")
    open(cfile, "w").write("content")

    responses.add(
        responses.POST,
        "https://host_name.com/customModels/",
        body=json.dumps(mocked_training_model),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.POST,
        "https://host_name.com/customModels/{}/versions/".format(mocked_training_model["id"]),
        body=json.dumps(mocked_custom_model_version),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.POST,
        make_blueprints_url(),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_blueprint),
    )
    responses.add(
        responses.GET,
        "https://host_name.com/customModels/{}/".format(mocked_training_model["id"]),
        body=json.dumps(mocked_training_model),
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        "https://host_name.com/executionEnvironments/{}/".format(mocked_environment["id"]),
        body=json.dumps(mocked_environment),
        status=200,
        content_type="application/json",
    )
    blueprint = CustomTrainingBlueprint.create_from_dropin(
        model_name="dave",
        dropin_env_id=mocked_environment["id"],
        target_type=TARGET_TYPE.REGRESSION,
        training_code_files=[bfile, cfile],
    )
    assert blueprint.id == mocked_blueprint["userBlueprintId"]
    assert blueprint.filenames == [bfile, cfile]



import trafaret as t

from datarobot.enums import FEATURE_TYPE
from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2, from_api


class FeatureEffectMetadata(APIObject):
    """ Feature Effect Metadata for model, contains status and available model sources.

    Notes
    -----

    `source` is expected parameter to retrieve Feature Fit. One of provided sources
    shall be used.

    """

    _converter = t.Dict(
        {t.Key("status"): t.String, t.Key("sources"): t.List(t.String)}
    ).ignore_extra("*")

    def __init__(self, status, sources):
        self.status = status
        self.sources = sources

    def __repr__(self):
        return encode_utf8_if_py2(u"FeatureEffectMetadata({}/{})".format(self.status, self.sources))


class FeatureEffectMetadataDatetime(APIObject):
    """ Feature Effect Metadata for datetime model, contains list of
    feature effect metadata per backtest.

    Notes
    -----
    ``feature effect metadata per backtest`` contains:
        * ``status`` : string.
        * ``backtest_index`` : string.
        * ``sources`` : list(string).

    `source` is expected parameter to retrieve Feature Fit. One of provided sources
    shall be used.

    `backtest_index` is expected parameter to submit compute request and retrieve Feature Effect.
    One of provided backtest indexes shall be used.

    Attributes
    ----------
    data : list[FeatureEffectMetadataDatetimePerBacktest]
        List feature effect metadata per backtest

    """

    _converter = t.Dict(
        {
            t.Key("data"): t.List(
                t.Dict(
                    {
                        t.Key("backtest_index"): t.String,
                        t.Key("status"): t.String,
                        t.Key("sources"): t.List(t.String),
                    }
                ).ignore_extra("*")
            )
        }
    ).ignore_extra("*")

    def __init__(self, data):
        self.data = [
            FeatureEffectMetadataDatetimePerBacktest(fe_meta_per_backtest)
            for fe_meta_per_backtest in data
        ]

    def __repr__(self):
        return encode_utf8_if_py2(u"FeatureEffectDatetimeMetadata({})".format(self.data))

    def __iter__(self):
        return iter(self.data)


class FeatureEffectMetadataDatetimePerBacktest(object):
    """Convert dictionary into feature effect metadata per backtest which contains backtest_index,
    status and sources.
    """

    def __init__(self, ff_metadata_datetime_per_backtest):
        self.backtest_index = ff_metadata_datetime_per_backtest["backtest_index"]
        self.status = ff_metadata_datetime_per_backtest["status"]
        self.sources = ff_metadata_datetime_per_backtest["sources"]

    def __repr__(self):
        return encode_utf8_if_py2(
            u"FeatureEffectMetadataDatetimePerBacktest(backtest_index={},"
            u"status={}, sources={}".format(self.backtest_index, self.status, self.sources)
        )

    def __eq__(self, other):
        return all(
            [
                self.backtest_index == other.backtest_index,
                self.status == other.status,
                sorted(self.sources) == sorted(other.sources),
            ]
        )

    def __lt__(self, other):
        return self.backtest_index < other.backtest_index


class FeatureEffects(APIObject):
    """
    Feature Effects provides partial dependence and predicted vs actual values for top-500
    features ordered by feature impact score.

    The partial dependence shows marginal effect of a feature on the target variable after
    accounting for the average effects of all other predictive features. It indicates how, holding
    all other variables except the feature of interest as they were, the value of this feature
    affects your prediction.

    Attributes
    ----------
    project_id: string
        The project that contains requested model
    model_id: string
        The model to retrieve Feature Effects for
    source: string
        The source to retrieve Feature Effects for
    feature_effects: list
        Feature Effects for every feature
    backtest_index: string, required only for DatetimeModels,
        The backtest index to retrieve Feature Effects for.

    Notes
    ------
    ``featureEffects`` is a dict containing the following:

        * ``feature_name`` (string) Name of the feature
        * ``feature_type`` (string) `dr.enums.FEATURE_TYPE`, \
          Feature type either numeric or categorical
        * ``feature_impact_score`` (float) Feature impact score
        * ``weight_label`` (string) optional, Weight label if configured for the project else null
        * ``partial_dependence`` (List) Partial dependence results
        * ``predicted_vs_actual`` (List) optional, Predicted versus actual results, \
          may be omitted if there are insufficient qualified samples

    ``partial_dependence`` is a dict containing the following:
        * ``is_capped`` (bool) Indicates whether the data for computation is capped
        * ``data`` (List) partial dependence results in the following format

    ``data`` is a list of dict containing the following:
        * ``label`` (string) Contains label for categorical and numeric features as string
        * ``dependence`` (float) Value of partial dependence

    ``predicted_vs_actual`` is a dict containing the following:
        * ``is_capped`` (bool) Indicates whether the data for computation is capped
        * ``data`` (List) pred vs actual results in the following format

    ``data`` is a list of dict containing the following:
        * ``label`` (string) Contains label for categorical features \
          for numeric features contains range or numeric value.
        * ``bin`` (List) optional, For numeric features contains \
          labels for left and right bin limits
        * ``predicted`` (float) Predicted value
        * ``actual`` (float) Actual value. Actual value is null \
          for unsupervised timeseries models
        * ``row_count`` (int or float) Number of rows for the label and bin. \
          Type is float if weight or exposure is set for the project.
    """

    _PartialDependence = t.Dict(
        {
            t.Key("is_capped"): t.Bool,
            t.Key("data"): t.List(
                t.Dict(
                    {t.Key("label"): t.String | t.Int, t.Key("dependence"): t.Float}
                ).ignore_extra("*")
            ),
        }
    ).ignore_extra("*")

    _PredictedVsActual = t.Dict(
        {
            t.Key("is_capped"): t.Bool,
            t.Key("data"): t.List(
                t.Dict(
                    {
                        t.Key("row_count"): t.Int | t.Float,
                        t.Key("label"): t.String,
                        t.Key("bin", optional=True): t.List(t.String),
                        t.Key("predicted"): t.Float | t.Null,
                        t.Key("actual"): t.Float | t.Null,
                    }
                ).ignore_extra("*")
            ),
        }
    ).ignore_extra("*")

    _converter = t.Dict(
        {
            t.Key("project_id"): t.String,
            t.Key("model_id"): t.String,
            t.Key("source"): t.String,
            t.Key("backtest_index", optional=True): t.String,
            t.Key("feature_effects"): t.List(
                t.Dict(
                    {
                        t.Key("feature_name"): t.String,
                        t.Key("feature_impact_score"): t.Float,
                        t.Key("feature_type"): t.Enum(
                            FEATURE_TYPE.NUMERIC, FEATURE_TYPE.CATEGORICAL
                        ),
                        t.Key("partial_dependence"): _PartialDependence,
                        t.Key("predicted_vs_actual", optional=True): _PredictedVsActual,
                        t.Key("weight_label", optional=True): t.String | t.Null,
                    }
                ).ignore_extra("*")
            ),
        }
    ).ignore_extra("*")

    def __init__(
        self, project_id, model_id, source, feature_effects, backtest_index=None,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.source = source
        self.backtest_index = backtest_index
        self.feature_effects = feature_effects

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}(project_id={},"
            u"model_id={}, source={}, backtest_index={}, feature_effects={}".format(
                self.__class__.__name__,
                self.project_id,
                self.model_id,
                self.source,
                self.backtest_index,
                self.feature_effects,
            )
        )

    def __eq__(self, other):
        return all(
            [
                self.project_id == other.project_id,
                self.model_id == other.model_id,
                self.source == other.source,
                self.backtest_index == other.backtest_index,
                (
                    sorted(self.feature_effects, key=lambda k: k["feature_name"])
                    == sorted(other.feature_effects, key=lambda k: k["feature_name"])
                ),
            ]
        )

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.project_id,
                self.model_id,
                self.source,
                self.backtest_index,
            )
        )

    def __iter__(self):
        return iter(self.feature_effects)

    @classmethod
    def from_server_data(cls, data):
        """
        Instantiate an object of this class using the data directly from the server,
        meaning that the keys may have the wrong camel casing.

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        """
        # keep_null_keys is required for predicted/actual
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)



from mock import patch
import pytest
import responses

from datarobot import ExternalScores
from datarobot.errors import ClientError


class TestExternalScores(object):
    @pytest.fixture()
    def scores_data(self, project_id, model_id):
        data = {
            "projectId": project_id,
            "modelId": model_id,
            "scores": [
                {"label": "AUC", "value": 1.0},
                {"label": "RMSE", "value": 0.0606},
                {"label": "Kolmogorov-Smirnov", "value": 1.0},
                {"label": "LogLoss", "value": 0.01383},
                {"label": "FVE Binomial", "value": -13845778590722.23},
                {"label": "Max MCC", "value": 1.0},
            ],
        }

        return data

    @staticmethod
    def create_response_data(count, scores, model_id=None, actual_value_column=None):
        data = []
        for i in range(count):
            item = dict(
                scores, datasetId="dataset_id_{}".format(i), actualValueColumn=actual_value_column
            )
            if model_id:
                item["modelId"] = model_id
            data.append(item)
        return data

    @pytest.fixture
    def url(self, project_id):
        return "https://host_name.com/projects/{}/externalScores/".format(project_id)

    def test_instantiation(self, scores_data):
        scores = self.create_response_data(1, scores_data, actual_value_column="label")[0]
        scores_obj = ExternalScores.from_server_data(scores)

        assert scores_obj.dataset_id == "dataset_id_0"
        assert scores_obj.project_id == scores_data["projectId"]
        assert scores_obj.model_id == scores_data["modelId"]
        assert scores_obj.actual_value_column == "label"
        assert scores_obj.scores == scores_data["scores"]

    def test_future_proof(self, scores_data):
        scores = self.create_response_data(1, scores_data)[0]
        data_with_future_keys = dict(scores, new_key="some future lift data")
        ExternalScores.from_server_data(data_with_future_keys)

    @responses.activate
    def test_scores_insights_create(self, project_id, project_url, base_job_server_data):
        scores_base_url = "{}externalScores/".format(project_url)
        job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])

        responses.add(
            responses.POST,
            scores_base_url,
            body="",
            status=202,
            content_type="application/json",
            adding_headers={"Location": job_url},
        )
        responses.add(
            responses.GET,
            job_url,
            json=base_job_server_data,
            status=303,
            adding_headers={"Location": job_url},
        )
        insights_job = ExternalScores.create(
            project_id=project_id,
            model_id="578e59a41ced2e5a9eb18960",
            dataset_id="581c6256100d2b60586980d9",
        )
        assert str(insights_job.id) == base_job_server_data["id"]

    @responses.activate
    def test_scores_insights_create_empty_target_unsupervised(
        self, project_id, project_url, base_job_server_data
    ):
        scores_base_url = "{}externalScores/".format(project_url)
        job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])

        responses.add(
            responses.POST,
            scores_base_url,
            body="",
            status=202,
            content_type="application/json",
            adding_headers={"Location": job_url},
        )
        exception_text = (
            "Cannot compute scores and insights on this dataset because target column is missing"
        )
        responses.add(
            responses.GET,
            job_url,
            json={"message": exception_text},
            status=422,
            adding_headers={"Location": job_url},
        )
        with pytest.raises(ClientError) as e:
            ExternalScores.create(
                project_id=project_id,
                model_id="578e59a41ced2e5a9eb18960",
                dataset_id="581c6256100d2b60586980d9",
            )
        assert exception_text in str(e.value)

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights.external_scores.DEFAULT_BATCH_SIZE", 10
    )
    def test_external_scores_all_results(self, scores_data, url, project_id, model_id):
        url_page1 = url + "?modelId={}&limit=10&offset=0".format(model_id)
        url_page2 = url + "?modelId={}&limit=10&offset=10".format(model_id)
        all_data = self.create_response_data(11, scores_data)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[10:])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True
        )

        scores_list = ExternalScores.list(project_id, model_id, limit=0)
        assert len(scores_list) == 11
        for i, scores in enumerate(scores_list):
            assert isinstance(scores, ExternalScores)
            assert scores.project_id == project_id
            assert scores.model_id == model_id
            assert scores.scores == scores_data["scores"]

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights.external_scores.DEFAULT_BATCH_SIZE", 1
    )
    def test_external_scores_all_results_with_offset(self, scores_data, url, project_id):
        url_page1 = url + "?datasetId={}&limit=1&offset=1".format("dataset_id_0")
        url_page2 = url + "?datasetId={}&limit=1&offset=2".format("dataset_id_0")
        all_data = self.create_response_data(1, scores_data, model_id="model_id_0")
        all_data.extend(self.create_response_data(1, scores_data, model_id="model_id_1"))
        all_data.extend(self.create_response_data(1, scores_data, model_id="model_id_2"))
        resp_data_1 = dict(count=1, previous=None, next=url_page2, data=[all_data[1]])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=[all_data[2]])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True
        )

        scores_list = ExternalScores.list(project_id, dataset_id="dataset_id_0", limit=0, offset=1)
        assert len(scores_list) == 2
        assert scores_list[0].dataset_id == "dataset_id_0"
        assert scores_list[0].model_id == "model_id_1"
        assert scores_list[1].dataset_id == "dataset_id_0"
        assert scores_list[1].model_id == "model_id_2"

    @responses.activate
    def test_external_scores_limited(self, scores_data, url, project_id, model_id):
        url_page = url + "?limit=3&offset=2"
        all_data = self.create_response_data(5, scores_data)
        resp_data = dict(count=3, previous=None, next=None, data=all_data[2:5])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        scores_list = ExternalScores.list(project_id, limit=3, offset=2)
        assert len(scores_list) == 3
        for i, scores in enumerate(scores_list):
            assert isinstance(scores, ExternalScores)
            assert scores.dataset_id == "dataset_id_{}".format(i + 2)

    @responses.activate
    def test_get_scores_success(self, scores_data, url, project_id, model_id):
        dataset_id = "dataset_id_0"
        url_page = url + "?modelId={}&datasetId={}&limit=100&offset=0".format(model_id, dataset_id)
        all_data = self.create_response_data(2, scores_data, model_id=model_id)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        scores = ExternalScores.get(project_id, model_id, dataset_id)
        assert isinstance(scores, ExternalScores)
        assert scores.dataset_id == dataset_id
        assert scores.model_id == model_id

    @responses.activate
    def test_get_scores_failure(self, url, project_id, model_id):
        dataset_id = "dataset_id_0"
        url_page = url + "?modelId={}&datasetId={}&limit=100&offset=0".format(model_id, dataset_id)
        resp_data = dict(count=0, previous=None, next=None, data=[])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        with pytest.raises(ValueError, match="dataset_id must be specified"):
            ExternalScores.get(project_id, model_id, None)
        with pytest.raises(ValueError, match="model_id must be specified"):
            ExternalScores.get(project_id, None, None)
        with pytest.raises(ClientError, match="Requested scores do not exist."):
            ExternalScores.get(project_id, model_id, dataset_id)



# -*- coding: utf-8 -*-
from itertools import product
import json

import pytest
import responses

from datarobot import Model
from datarobot.errors import ClientError
from datarobot.models.roc_curve import RocCurve
from datarobot.utils import from_api


@pytest.fixture
def roc_curve_validation_data(model_id, roc_curve_data):
    return {
        "source": "validation",
        "sourceModelId": model_id,
        "rocPoints": roc_curve_data["rocPoints"],
        "negativeClassPredictions": roc_curve_data["negativeClassPredictions"],
        "positiveClassPredictions": roc_curve_data["positiveClassPredictions"],
    }


@pytest.fixture
def roc_curve_parent_model_validation_data(parent_model_id, roc_curve_data):
    return {
        "source": "validation",
        "sourceModelId": parent_model_id,
        "rocPoints": roc_curve_data["rocPoints"],
        "negativeClassPredictions": roc_curve_data["negativeClassPredictions"],
        "positiveClassPredictions": roc_curve_data["positiveClassPredictions"],
    }


@pytest.fixture
def roc_curve_parent_model_holdout_data(parent_model_id, roc_curve_data):
    return {
        "source": "holdout",
        "sourceModelId": parent_model_id,
        "rocPoints": roc_curve_data["rocPoints"],
        "negativeClassPredictions": roc_curve_data["negativeClassPredictions"],
        "positiveClassPredictions": roc_curve_data["positiveClassPredictions"],
    }


def test_instantiation(roc_curve_validation_data):
    roc = RocCurve.from_server_data(roc_curve_validation_data)

    assert roc.source == roc_curve_validation_data["source"]
    assert roc.negative_class_predictions == roc_curve_validation_data["negativeClassPredictions"]
    assert roc.positive_class_predictions == roc_curve_validation_data["positiveClassPredictions"]
    assert roc.roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc.source_model_id == roc_curve_validation_data["sourceModelId"]


def test_future_proof(roc_curve_validation_data):
    data_with_future_keys = dict(roc_curve_validation_data, new_key="some future roc data")
    data_with_future_keys["rocPoints"][0]["new_key"] = "some future bin data"
    RocCurve.from_server_data(data_with_future_keys)


@pytest.fixture
def roc_curve_validation_data_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/rocCurve/validation/".format(
        project_id, model_id
    )


@pytest.fixture
def roc_curve_parent_model_validation_data_url(project_id, parent_model_id):
    return "https://host_name.com/projects/{}/models/{}/rocCurve/validation/".format(
        project_id, parent_model_id
    )


@pytest.fixture
def roc_curve_list_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/rocCurve/".format(project_id, model_id)


@pytest.fixture
def roc_curve_parent_list_url(project_id, parent_model_id):
    return "https://host_name.com/projects/{}/models/{}/rocCurve/".format(
        project_id, parent_model_id
    )


@pytest.fixture
def frozen_model_url(project_id, model_id):
    return "https://host_name.com/projects/{}/frozenModels/{}/".format(project_id, model_id)


@responses.activate
def test_get_validation_roc_curve(
    roc_curve_validation_data, roc_curve_validation_data_url, project_id, model_id
):
    responses.add(
        responses.GET,
        roc_curve_validation_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(roc_curve_validation_data),
    )
    model = Model(id=model_id, project_id=project_id)
    roc = model.get_roc_curve("validation")

    assert roc.source == roc_curve_validation_data["source"]
    assert roc.negative_class_predictions == roc_curve_validation_data["negativeClassPredictions"]
    assert roc.positive_class_predictions == roc_curve_validation_data["positiveClassPredictions"]
    assert roc.roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc.source_model_id == model_id


@responses.activate
def test_get_frozen_validation_roc_curve_no_fallback(
    roc_curve_validation_data_url, project_id, model_id
):
    responses.add(responses.GET, roc_curve_validation_data_url, status=404)
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    with pytest.raises(ClientError):
        model.get_roc_curve("validation")


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_frozen_validation_roc_curve_with_fallback(
    roc_curve_parent_model_validation_data,
    roc_curve_validation_data_url,
    roc_curve_parent_model_validation_data_url,
    frozen_model_url,
    project_id,
    model_id,
    parent_model_id,
    frozen_json,
):
    responses.add(responses.GET, roc_curve_validation_data_url, status=404)
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        roc_curve_parent_model_validation_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(roc_curve_parent_model_validation_data),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc = model.get_roc_curve("validation", fallback_to_parent_insights=True)

    assert roc.source == roc_curve_parent_model_validation_data["source"]
    assert (
        roc.negative_class_predictions
        == roc_curve_parent_model_validation_data["negativeClassPredictions"]
    )
    assert (
        roc.positive_class_predictions
        == roc_curve_parent_model_validation_data["positiveClassPredictions"]
    )
    assert roc.roc_points == from_api(roc_curve_parent_model_validation_data["rocPoints"])
    assert roc.source_model_id == parent_model_id


@responses.activate
def test_get_all_roc_curves(roc_curve_validation_data, roc_curve_list_url, project_id, model_id):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [roc_curve_validation_data]}),
    )
    model = Model(id=model_id, project_id=project_id)
    roc_list = model.get_all_roc_curves()

    assert len(roc_list) == 1
    assert roc_list[0].source == roc_curve_validation_data["source"]
    assert (
        roc_list[0].negative_class_predictions
        == roc_curve_validation_data["negativeClassPredictions"]
    )
    assert (
        roc_list[0].positive_class_predictions
        == roc_curve_validation_data["positiveClassPredictions"]
    )
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc_list[0].source_model_id == model_id


@responses.activate
def test_get_frozen_all_roc_curves_no_fallback(
    roc_curve_validation_data, roc_curve_list_url, project_id, model_id
):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [roc_curve_validation_data]}),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc_list = model.get_all_roc_curves()

    assert len(roc_list) == 1
    assert roc_list[0].source == roc_curve_validation_data["source"]
    assert (
        roc_list[0].negative_class_predictions
        == roc_curve_validation_data["negativeClassPredictions"]
    )
    assert (
        roc_list[0].positive_class_predictions
        == roc_curve_validation_data["positiveClassPredictions"]
    )
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc_list[0].source_model_id == model_id


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_forzen_all_roc_curves_with_fallback(
    roc_curve_validation_data,
    roc_curve_parent_model_validation_data,
    roc_curve_parent_model_holdout_data,
    roc_curve_list_url,
    roc_curve_parent_list_url,
    frozen_model_url,
    project_id,
    model_id,
    parent_model_id,
    frozen_json,
):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [roc_curve_validation_data]}),
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        roc_curve_parent_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps(
            {
                "charts": [
                    roc_curve_parent_model_validation_data,
                    roc_curve_parent_model_holdout_data,
                ]
            }
        ),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc_list = model.get_all_roc_curves(fallback_to_parent_insights=True)

    assert len(roc_list) == 2
    assert roc_list[0].source == roc_curve_validation_data["source"]
    assert (
        roc_list[0].negative_class_predictions
        == roc_curve_validation_data["negativeClassPredictions"]
    )
    assert (
        roc_list[0].positive_class_predictions
        == roc_curve_validation_data["positiveClassPredictions"]
    )
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc_list[0].source_model_id == model_id

    assert roc_list[1].source == roc_curve_parent_model_holdout_data["source"]
    assert (
        roc_list[1].negative_class_predictions
        == roc_curve_parent_model_holdout_data["negativeClassPredictions"]
    )
    assert (
        roc_list[1].positive_class_predictions
        == roc_curve_parent_model_holdout_data["positiveClassPredictions"]
    )
    assert roc_list[1].roc_points == from_api(roc_curve_parent_model_holdout_data["rocPoints"])
    assert roc_list[1].source_model_id == parent_model_id


def test_get_best_f1_threshold(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When calculate recommended threshold
    best_threshold = roc.get_best_f1_threshold()
    # Then f1 score for that threshold maximal from all ROC curve points
    best_f1 = roc.estimate_threshold(best_threshold)["f1_score"]
    assert all(best_f1 >= roc_point["f1_score"] for roc_point in roc.roc_points)


def test_estimate_threshold_equal(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When estimating threshold equal to one of precalculated
    threshold = roc_curve_validation_data["rocPoints"][1]["threshold"]
    # Then estimate_threshold return valid data
    assert roc.estimate_threshold(threshold)["threshold"] == threshold


def test_estimate_threshold_new(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When estimating threshold from outside of precalculated
    threshold = roc_curve_validation_data["rocPoints"][1]["threshold"] + 0.1
    # Then estimate_threshold return data for next threshold bigger then requested
    assert roc.estimate_threshold(threshold)["threshold"] > threshold


class TestLabelwiseRocCurve:
    @pytest.fixture
    def labelwise_roc_curve_validation_data_url(self, project_id, model_id):
        return "https://host_name.com/projects/{}/models/{}/labelwiseRocCurves/validation/".format(
            project_id, model_id
        )

    @pytest.fixture
    def labelwise_roc_curve_parent_model_validation_data_url(self, project_id, parent_model_id):
        return "https://host_name.com/projects/{}/models/{}/labelwiseRocCurves/validation/".format(
            project_id, parent_model_id
        )

    @pytest.fixture
    def labelwise_roc_curve_list_url(self, project_id, model_id):
        return "https://host_name.com/projects/{}/models/{}/labelwiseRocCurves/".format(
            project_id, model_id
        )

    @pytest.fixture
    def labelwise_roc_curve_parent_list_url(self, project_id, parent_model_id):
        return "https://host_name.com/projects/{}/models/{}/labelwiseRocCurves/".format(
            project_id, parent_model_id
        )

    @pytest.fixture
    def labelwise_roc_curve_validation_data(self, roc_curve_data):
        return self.get_labelwise_roc_curves_response(["validation"], roc_curve_data)

    @pytest.fixture
    def labelwise_roc_curve_validation_and_cross_validation_data(self, roc_curve_data):
        return self.get_labelwise_roc_curves_response(
            ["validation", "crossValidation"], roc_curve_data
        )

    def get_labelwise_roc_curves_response(self, sources, roc_curve_data):
        LABELS = ["cat", u"🐕"]
        charts = []
        for label in LABELS:
            for source in sources:
                chart_dict = dict(label=label, source=source, kolmogorovSmirnovMetric=0.1, auc=0.2,)
                chart_dict.update(roc_curve_data)
                charts.append(chart_dict)
        response = {
            "labels": LABELS,
            "charts": charts,
            "rocType": "labelwise",
        }
        return response

    @pytest.fixture
    def labelwise_roc_curve_validation_cross_validation_and_holdout_data(self, roc_curve_data):
        return self.get_labelwise_roc_curves_response(
            ["validation", "crossValidation", "holdout"], roc_curve_data
        )

    @responses.activate
    def test_get_labelwise_roc_curves(
        self,
        labelwise_roc_curve_validation_data,
        labelwise_roc_curve_validation_data_url,
        project_id,
        model_id,
        roc_curve_data,
    ):
        responses.add(
            responses.GET,
            labelwise_roc_curve_validation_data_url,
            status=200,
            content_type="application/json",
            body=json.dumps(labelwise_roc_curve_validation_data),
        )
        model = Model(id=model_id, project_id=project_id)
        roc_charts = model.get_labelwise_roc_curves("validation")

        expected_labels = ["cat", u"🐕"]
        assert len(roc_charts) == len(expected_labels)

        for roc_chart, expected_label in zip(roc_charts, expected_labels):
            self._assert_single_labelwise_roc_curve(
                roc_chart, expected_label, "validation", model_id, roc_curve_data
            )

    def _assert_single_labelwise_roc_curve(
        self, roc_chart, expected_label, expected_source, expected_model_id, expected_roc_curve_data
    ):
        assert roc_chart.source == expected_source
        assert roc_chart.source_model_id == expected_model_id
        assert roc_chart.label == expected_label
        assert roc_chart.kolmogorov_smirnov_metric == 0.1
        assert roc_chart.auc == 0.2
        assert roc_chart.roc_points == from_api(expected_roc_curve_data["rocPoints"])
        assert (
            roc_chart.negative_class_predictions
            == expected_roc_curve_data["negativeClassPredictions"]
        )
        assert (
            roc_chart.positive_class_predictions
            == expected_roc_curve_data["positiveClassPredictions"]
        )

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_get_labelwise_roc_curves_with_fallback(
        self,
        labelwise_roc_curve_validation_data,
        labelwise_roc_curve_validation_data_url,
        labelwise_roc_curve_parent_model_validation_data_url,
        project_id,
        model_id,
        parent_model_id,
        roc_curve_data,
        frozen_json,
        frozen_model_url,
    ):
        responses.add(
            responses.GET, labelwise_roc_curve_validation_data_url, status=404,
        )

        responses.add(
            responses.GET,
            frozen_model_url,
            body=frozen_json,
            status=200,
            content_type="application/json",
        )
        responses.add(
            responses.GET,
            labelwise_roc_curve_parent_model_validation_data_url,
            status=200,
            content_type="application/json",
            body=json.dumps(labelwise_roc_curve_validation_data),
        )
        model = Model(id=model_id, project_id=project_id, is_frozen=True)
        roc_charts = model.get_labelwise_roc_curves("validation", fallback_to_parent_insights=True)

        expected_labels = ["cat", u"🐕"]
        assert len(roc_charts) == len(expected_labels)

        for roc_chart, expected_label in zip(roc_charts, expected_labels):
            self._assert_single_labelwise_roc_curve(
                roc_chart, expected_label, "validation", parent_model_id, roc_curve_data
            )

    @responses.activate
    def test_get_labelwise_roc_curves_no_data_available(
        self, labelwise_roc_curve_validation_data_url, project_id, model_id, parent_model_id,
    ):
        responses.add(responses.GET, labelwise_roc_curve_validation_data_url, status=404)
        model = Model(id=model_id, project_id=project_id)
        with pytest.raises(ClientError):
            model.get_labelwise_roc_curves("validation")

    @responses.activate
    def test_get_all_labelwise_roc_curves(
        self,
        labelwise_roc_curve_validation_and_cross_validation_data,
        labelwise_roc_curve_list_url,
        project_id,
        model_id,
        roc_curve_validation_data,
        roc_curve_data,
    ):
        responses.add(
            responses.GET,
            labelwise_roc_curve_list_url,
            status=200,
            content_type="application/json",
            body=json.dumps(labelwise_roc_curve_validation_and_cross_validation_data),
        )
        model = Model(id=model_id, project_id=project_id)
        roc_charts = model.get_all_labelwise_roc_curves()

        expected_labels = ["cat", u"🐕"]
        expected_sources = ["validation", "crossValidation"]
        expected_source_label_combinations = product(expected_labels, expected_sources)

        assert len(roc_charts) == len(expected_labels) * len(expected_sources)

        for roc_chart, (expected_label, expected_source) in zip(
            roc_charts, expected_source_label_combinations
        ):
            self._assert_single_labelwise_roc_curve(
                roc_chart, expected_label, expected_source, model_id, roc_curve_data
            )

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_get_all_labelwise_roc_curves_fallback_to_parent_insight(
        self,
        labelwise_roc_curve_validation_and_cross_validation_data,
        labelwise_roc_curve_validation_cross_validation_and_holdout_data,
        labelwise_roc_curve_list_url,
        labelwise_roc_curve_parent_list_url,
        project_id,
        model_id,
        parent_model_id,
        roc_curve_validation_data,
        frozen_model_url,
        frozen_json,
        roc_curve_data,
    ):
        responses.add(
            responses.GET,
            labelwise_roc_curve_list_url,
            status=200,
            content_type="application/json",
            body=json.dumps(labelwise_roc_curve_validation_and_cross_validation_data),
        )
        responses.add(
            responses.GET,
            frozen_model_url,
            body=frozen_json,
            status=200,
            content_type="application/json",
        )
        responses.add(
            responses.GET,
            labelwise_roc_curve_parent_list_url,
            status=200,
            content_type="application/json",
            body=json.dumps(labelwise_roc_curve_validation_cross_validation_and_holdout_data),
        )
        model = Model(id=model_id, project_id=project_id, is_frozen=True)
        roc_charts = model.get_all_labelwise_roc_curves(fallback_to_parent_insights=True)

        expected_labels = ["cat", u"🐕"]
        expected_sources = ["validation", "crossValidation", "holdout"]

        expected_source_label_combinations = [
            ("cat", "validation"),
            ("cat", "crossValidation"),
            (u"🐕", "validation"),
            (u"🐕", "crossValidation"),
            ("cat", "holdout"),
            (u"🐕", "holdout"),
        ]
        assert len(roc_charts) == len(expected_labels) * len(expected_sources)

        for roc_chart, (expected_label, expected_source) in zip(
            roc_charts, expected_source_label_combinations
        ):
            expected_model_id = parent_model_id if expected_source == "holdout" else model_id
            self._assert_single_labelwise_roc_curve(
                roc_chart, expected_label, expected_source, expected_model_id, roc_curve_data
            )

    @responses.activate
    def test_get_all_labelwise_roc_curves_fallback_flag_not_set(
        self,
        labelwise_roc_curve_validation_and_cross_validation_data,
        labelwise_roc_curve_validation_cross_validation_and_holdout_data,
        labelwise_roc_curve_list_url,
        labelwise_roc_curve_parent_list_url,
        project_id,
        model_id,
        parent_model_id,
        roc_curve_validation_data,
        frozen_model_url,
        frozen_json,
        roc_curve_data,
    ):
        responses.add(
            responses.GET,
            labelwise_roc_curve_list_url,
            status=200,
            content_type="application/json",
            body=json.dumps(labelwise_roc_curve_validation_and_cross_validation_data),
        )
        responses.add(
            responses.GET,
            frozen_model_url,
            body=frozen_json,
            status=200,
            content_type="application/json",
        )
        responses.add(
            responses.GET,
            labelwise_roc_curve_parent_list_url,
            status=200,
            content_type="application/json",
            body=json.dumps(labelwise_roc_curve_validation_cross_validation_and_holdout_data),
        )
        model = Model(id=model_id, project_id=project_id, is_frozen=True)
        roc_charts = model.get_all_labelwise_roc_curves(fallback_to_parent_insights=False)

        expected_labels = ["cat", u"🐕"]
        expected_sources = ["validation", "crossValidation"]
        expected_source_label_combinations = product(expected_labels, expected_sources)

        for roc_chart, (expected_label, expected_source) in zip(
            roc_charts, expected_source_label_combinations
        ):
            expected_model_id = parent_model_id if expected_source == "holdout" else model_id
            self._assert_single_labelwise_roc_curve(
                roc_chart, expected_label, expected_source, expected_model_id, roc_curve_data
            )



import datetime
import unittest

import pandas as pd
import pytest
import responses

from datarobot import errors
from datarobot.client import Client, set_client
from datarobot.errors import DataRobotDeprecationWarning
from datarobot.utils import (
    dataframe_to_buffer,
    deprecation,
    from_api,
    get_id_from_response,
    is_urlsource,
    parse_time,
    recognize_sourcedata,
    to_api,
)
from datarobot.utils.sourcedata import list_of_records_to_buffer
from tests.test_helpers import fixture_file_path
from tests.utils import warns


class TestDataframeSerialization(unittest.TestCase):
    def test_list_of_records_to_buffer(self):
        records = [{"a": "b", "c": "d"}, {"a": "x", "c": "y"}]
        expected = "a,c\r\nb,d\r\nx,y\r\n"
        buff = list_of_records_to_buffer(records)
        self.assertEqual(buff.read(), expected)

    def test_no_index_please(self):
        df = pd.DataFrame({"a": range(100), "b": range(100)})
        buff = dataframe_to_buffer(df)
        readback = pd.read_csv(buff)
        self.assertEqual(readback.columns.tolist(), ["a", "b"])

    def test_parse_time(self):
        self.assertEqual("BAD TIME", parse_time("BAD TIME"))  # returns value
        test_string_time = datetime.datetime.now().isoformat()
        self.assertIsInstance(test_string_time, str)
        self.assertIsInstance(parse_time(test_string_time), datetime.datetime)

    @responses.activate
    @pytest.mark.usefixtures("no_client_version_check")
    def test_get_id_from_response_location_header(self):
        responses.add(
            responses.POST,
            "http://nothing/",
            body="",
            adding_headers={"Location": "http://nothing/f-id/"},
        )
        client = set_client(Client(token="no_matter", endpoint="http://nothing"))
        resp = client.post("")
        self.assertEqual(get_id_from_response(resp), "f-id")

    def test_csv_buffer_is_quoted(self):
        """"
        This tests that dataframes are correctly quoted when saved back to csv, so
        that columns containing line endings won't be interpreted as multiple rows
        See: PRED-3203
        """
        df = pd.DataFrame({"A": ["test\rtest", 2], "B": ["test\n\n\n\r\ntest", 0.75]})
        buff = dataframe_to_buffer(df)
        readback = pd.read_csv(buff)
        assert df.shape[0] == readback.shape[0]


class TestFromAPI(unittest.TestCase):
    def test_nested_list_of_objects_all_changed(self):
        source = {"oneFish": [{"twoFish": "redFish"}, {"blueFish": "noFish"}]}
        result = from_api(source)
        inner = result["one_fish"]
        self.assertEqual(inner[0]["two_fish"], "redFish")
        self.assertEqual(inner[1]["blue_fish"], "noFish")

    def test_nested_objects_all_changed(self):
        source = {"oneFish": {"twoFish": "redFish"}}

        result = from_api(source)
        self.assertEqual(result["one_fish"]["two_fish"], "redFish")


def test_from_api_filter_nones():
    data_in = {
        "someValue": {"nestedNone": None, "nestedNotNone": ""},
        "listNone": [{"someValue": None, "someNotNone": ""}],
        "otherNone": None,
    }
    data_out = {"some_value": {"nested_not_none": ""}, "list_none": [{"some_not_none": ""}]}
    assert data_out == from_api(data_in)


def test_from_api_no_filter_nones():
    data_in = {
        "someValue": {"nestedNone": None, "nestedNotNone": ""},
        "listNone": [{"someValue": None, "someNotNone": ""}],
        "otherNone": None,
    }
    data_out = {
        "some_value": {"nested_none": None, "nested_not_none": ""},
        "list_none": [{"some_value": None, "some_not_none": ""}],
        "other_none": None,
    }
    assert data_out == from_api(data_in, keep_null_keys=True)


def test_from_api_keep_nones():
    data_in = {
        "someValue": {"nestedNone": None, "nestedNotNone": "", "levelTwo": {"levelThree": None}},
        "otherNone": None,
        "keepLevelOne": None,
    }
    data_out = {
        "some_value": {
            "nested_not_none": "",
            "nested_none": None,
            "level_two": {"level_three": None},
        },
        "keep_level_one": None,
    }
    assert data_out == from_api(
        data_in,
        keep_attrs=["some_value.nested_none", "some_value.level_two.level_three", "keep_level_one"],
    )


def test_from_api_multiple_underscore_ignored():
    in_data = {"samplePct__gte": 64}
    expected = {"sample_pct__gte": 64}
    assert from_api(in_data) == expected


def test_to_api_multiple_underscore_ignored():
    in_data = {"sample_pct__gte": 64}
    expected = {"samplePct__gte": 64}
    assert to_api(in_data) == expected


def test_from_api_single_letter():
    in_data = {"iAmAKey": "test"}
    expected = {"i_am_a_key": "test"}
    assert from_api(in_data) == expected


def test_to_api_single_letter():
    in_data = {"i_am_a_key": "test"}
    expected = {"iAmAKey": "test"}
    assert to_api(in_data) == expected


@deprecation.deprecated(deprecated_since_version="v0.1.2", will_remove_version="v1.2.3")
def bar(arg1, arg2):
    """
    A dummy function to use for testing the deprecation mechanisms

    Parameters
    ----------
    arg1 : int
    arg2 : int

    Returns
    -------
    total : int
        arg1 + arg2

    """
    return arg1 + arg2


@deprecation.deprecated(
    deprecated_since_version="v0.1.2",
    will_remove_version="v1.2.3",
    message="Use of `bar` is recommended instead.",
)
def foo(arg1, arg2):
    """
    A dummy function to use for testing the deprecation mechanisms

    Parameters
    ----------
    arg1 : int
    arg2 : int

    Returns
    -------
    total : int
        foo + bar

    """
    return arg1 + arg2


def test_deprecation_with_message(known_warning):
    foo(1, 2)
    w_msg = known_warning[0].message.args[0]
    assert (
        w_msg == "`foo` has been deprecated in `v0.1.2`, will be removed "
        "in `v1.2.3`. Use of `bar` is recommended instead."
    )


def test_deprecation_message(known_warning):
    bar(1, 2)
    w_instance = known_warning[0].message
    assert isinstance(w_instance, DataRobotDeprecationWarning)


class TestSourcedataUtils(unittest.TestCase):
    def setUp(self):
        self.default_fname = "predict.csv"

    def test_recognize_sourcedata_passed_dataframe(self):
        df = pd.DataFrame({"a": range(100), "b": range(100)})
        kwargs = recognize_sourcedata(df, self.default_fname)
        self.assertTrue("filelike" in kwargs)
        self.assertEqual(kwargs.get("fname"), self.default_fname)
        buff = kwargs["filelike"]
        readback = pd.read_csv(buff)
        self.assertEqual(readback.columns.tolist(), ["a", "b"])

    def test_recognize_sourcedata_passed_filelike(self):
        path = fixture_file_path("synthetic-100.csv")
        with open(path, "rb") as fd:
            kwargs = recognize_sourcedata(fd, self.default_fname)
            self.assertTrue(kwargs.get("filelike") is fd)
            self.assertEqual(kwargs.get("fname"), self.default_fname)

    def test_recognize_sourcedata_passed_filepath(self):
        file_path = fixture_file_path("synthetic-100.csv")
        kwargs = recognize_sourcedata(file_path, self.default_fname)
        self.assertEqual(kwargs.get("file_path"), file_path)
        self.assertEqual(kwargs.get("fname"), "synthetic-100.csv")

    def test_recognize_sourcedata_passed_content(self):
        content = b"a,b,c\n" + b"1,2,3\n" * 100
        kwargs = recognize_sourcedata(content, self.default_fname)
        self.assertEqual(kwargs.get("content"), content)
        self.assertEqual(kwargs.get("fname"), self.default_fname)

    def test_is_urlsource_passed_true(self):
        result = is_urlsource("http://path_to_urlsource")
        self.assertTrue(result)

    def test_is_urlsource_file_passed_true(self):
        result = is_urlsource("file://path_to_urlsource")
        self.assertTrue(result)

    def test_is_urlsource_file_passed_false(self):
        result = is_urlsource("filename")
        self.assertFalse(result)

    def test_is_urlsource_passed_false(self):
        result = is_urlsource("not_a_path_to_urlsource")
        self.assertFalse(result)

    def test_is_urlsource_s3_passed_true(self):
        result = is_urlsource("s3://path/to/a/bucket/data.csv")
        self.assertTrue(result)

    def test_fatfingered_filepath_raises(self):
        content = b"/home/datarobot/mistypefilepath.csv"
        with self.assertRaises(errors.InputNotUnderstoodError):
            recognize_sourcedata(content, self.default_fname)


def test_resource():
    """Python 3 ResourceWarning of implicitly closed files."""
    with warns():
        open(__file__)


class TestImports(unittest.TestCase):
    def test_rename(self):
        """datarobot_sdk rename."""
        with self.assertRaises(ImportError):
            import datarobot_sdk  # noqa



#! /usr/bin/env python3
"""Show sample images for a project.

The following will open a project, get a list of sample images, and
then display a few images to the GUI.

The parameters may be adjusted to use your project name, feature name, and
the number of images to display.
"""
import io

import PIL.Image

from datarobot.models import Project
from datarobot.models.visualai import SampleImage


def display_images(project_name, feature_name, max_images):
    project = Project.list(search_params={"project_name": project_name})[0]
    for sample in SampleImage.list(project.id, feature_name)[:max_images]:
        with io.BytesIO(sample.image.image_bytes) as bio, PIL.Image.open(bio) as img:
            img.show()


if __name__ == "__main__":
    project_name = "dataset_2k.zip"
    feature_name = "image"
    max_images = 2
    display_images(project_name, feature_name, max_images)



import pytest
import trafaret as t
from trafaret import DataError

from datarobot.models.api_object import APIObject


def test_safe_data_renaming():
    class Dummy(APIObject):
        _converter = t.Dict({t.Key("renamed") >> "magic": t.String()})

    data = Dummy._safe_data({"renamed": "nombre"})
    assert data == {"magic": "nombre"}


@pytest.mark.parametrize(
    "data, do_recursive, validation_success",
    [
        ({"l1Prop": {"l2Prop": "x"}}, True, True),
        ({"l1Prop": {"l2_prop": "x"}}, True, True),
        ({"l1Prop": {"l2_prop": "x"}}, False, True),
        ({"l1Prop": {"l2Prop": "x"}}, False, False),
    ],
    ids=[
        "success-converted-recursive",
        "success-convertion_not_needed-recursive",
        "success-convertion_not_needed-not_recursive",
        "failure-no_convertion-not_recursive",
    ],
)
def test_safe_data_recursive(data, do_recursive, validation_success):
    class Dummy(APIObject):
        _converter = t.Dict({t.Key("l1_prop"): t.Dict({t.Key("l2_prop"): t.String()})})

    if validation_success:
        safe_data = Dummy._safe_data(data, do_recursive=do_recursive)
        assert safe_data == {"l1_prop": {"l2_prop": "x"}}
    else:
        with pytest.raises(DataError):
            Dummy._safe_data(data, do_recursive=do_recursive)



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2

LiftChartBinsTrafaret = t.Dict(
    {
        t.Key("bins"): t.List(
            t.Dict(
                {
                    t.Key("actual"): t.Float,
                    t.Key("predicted"): t.Float,
                    t.Key("bin_weight"): t.Float,
                }
            ).ignore_extra("*")
        )
    }
)


class LiftChart(APIObject):
    """ Lift chart data for model.

    Notes
    -----
    ``LiftChartBin`` is a dict containing the following:

        * ``actual`` (float) Sum of actual target values in bin
        * ``predicted`` (float) Sum of predicted target values in bin
        * ``bin_weight`` (float) The weight of the bin. For weighted projects, it is the sum of \
          the weights of the rows in the bin. For unweighted projects, it is the number of rows in \
          the bin.

    Attributes
    ----------
    source : str
        Lift chart data source. Can be 'validation', 'crossValidation' or 'holdout'.
    bins : list of dict
        List of dicts with schema described as ``LiftChartBin`` above.
    source_model_id : str
        ID of the model this lift chart represents; in some cases,
        insights from the parent of a frozen model may be used
    target_class : str, optional
        For multiclass lift - target class for this lift chart data.
    """

    _converter = (
        t.Dict(
            {
                t.Key("source"): t.String,
                t.Key("source_model_id"): t.String,
                t.Key("target_class", optional=True, default=None): t.String | t.Null,
            }
        )
        .merge(LiftChartBinsTrafaret)
        .ignore_extra("*")
    )

    def __init__(self, source, bins, source_model_id, target_class):
        self.source = source
        self.bins = bins
        self.source_model_id = source_model_id
        self.target_class = target_class

    def __repr__(self):
        if self.target_class:
            return encode_utf8_if_py2(u"LiftChart({}:{})".format(self.target_class, self.source))
        return encode_utf8_if_py2(u"LiftChart({})".format(self.source))
